Package: saber
Type: Package
Title: Context Engineering for Large Language Model Agents
Version: 0.7.2
Date: 2026-06-24
Authors@R: c(
    person("Troy", "Hernandez", role = c("aut", "cre"),
           email = "troy@cornball.ai",
           comment = c(ORCID = "0009-0005-4248-604X")),
    person("cornball.ai", role = "cph"))
Description: Context-engineering primitives for Artificial Intelligence (AI)
    coding agents working in R. Assembles agent context from memory and
    instruction files ('AGENTS.md', 'CLAUDE.md'), traces function call
    blast radius across projects, generates project briefings, parses
    source into Abstract Syntax Tree (AST) symbol indices, discovers
    dependency graphs, and introspects installed packages.
    Zero dependencies.
License: Apache License (>= 2)
Depends: R (>= 4.4.0)
URL: https://github.com/cornball-ai/saber
BugReports: https://github.com/cornball-ai/saber/issues
Suggests: tinytest
Encoding: UTF-8
RoxygenNote: 7.3.2
NeedsCompilation: no
Packaged: 2026-06-24 19:31:14 UTC; troy
Author: Troy Hernandez [aut, cre] (ORCID:
    <https://orcid.org/0009-0005-4248-604X>),
  cornball.ai [cph]
Maintainer: Troy Hernandez <troy@cornball.ai>
Repository: CRAN
Date/Publication: 2026-06-25 05:20:02 UTC
Built: R 4.4.3; ; 2026-06-25 05:56:49 UTC; unix

# tinyrox says don't edit this manually, but it can't stop you!

export(agent_context)
export(blast_radius)
export(briefing)
export(default_exclude)
export(find_downstream)
export(fn_graph)
export(graph_svg)
export(pkg_exports)
export(pkg_graph)
export(pkg_help)
export(pkg_internals)
export(projects)
export(symbols)

importFrom(stats,runif)
importFrom(utils,getParseData)
importFrom(utils,head)

# saber

**Context engineering for R.**

saber ("to know" in Spanish, pronounced [sah-BEHR](https://www.youtube.com/watch?v=m3WBocsw9lw)) assembles agent context, traces blast radius across projects, and introspects packages so AI coding agents don't have to guess.

## Install

```r
install.packages("saber")
# or to install the development version
remotes::install_github("cornball-ai/saber")
```

## Running these examples

Examples below use `Rscript -e` for portability (Linux, macOS, Windows). On *nix (Linux and macOS), [littler](https://eddelbuettel.github.io/littler/) (`r`) gives faster startup, but `r -e` does not auto-print return values. These three are equivalent:

```bash
Rscript -e 'saber::pkg_exports("saber")'         # portable
r -p -e 'saber::pkg_exports("saber")'            # littler, auto-print
r -e 'print(saber::pkg_exports("saber"))'        # littler, explicit print
```

See the [tinyverse development toolchain](https://cornball.ai/posts/tinyverse-development-toolchain/) for the full setup.

## What it does

**13 exported functions.**

### Agent context

| Function | What it does |
|---|---|
| `agent_context()` | Assemble memory, identity, and instruction files for an agent |
| `briefing()` | Generate a project briefing (metadata, dependents, git log) |

### Code intelligence

| Function | What it does |
|---|---|
| `symbols()` | Parse R source into function defs and calls via `getParseData()` |
| `blast_radius()` | Find every caller of a function, across projects |
| `fn_graph()` | Render a project's internal function call graph as SVG |
| `pkg_graph()` | Render a package dependency graph as SVG |
| `graph_svg()` | Force-directed graph renderer (used by `fn_graph` and `pkg_graph`) |

### Project discovery

| Function | What it does |
|---|---|
| `projects()` | Discover R package projects and their metadata |
| `find_downstream()` | Find all projects that depend on a given package |
| `default_exclude()` | Default directories to skip when scanning |

### Package introspection

| Function | What it does |
|---|---|
| `pkg_exports()` | List exported functions with argument signatures |
| `pkg_internals()` | List internal (non-exported) functions |
| `pkg_help()` | Pull help documentation as markdown |

## Examples

Assemble agent context from project and workspace files:

```r
# Claude Code agent in current project
saber::agent_context(agent = "claude")

# Codex agent with workspace identity
saber::agent_context(agent = "codex", workspace_dir = "~/.codex/workspace")
```

Generate a project briefing:

```r
saber::briefing("saber")
#> # Briefing: saber
#> _Generated 2026-03-25 00:30_
#>
#> ## Package
#> - **Name**: saber
#> - **Title**: Context Engineering for R
#> - **Version**: 0.7.0
#>
#> ## Recent commits
#> - 7983478 Add r-ci GitHub Actions workflow
#> - ...
```

Index all function definitions and calls in a project:

```r
syms <- saber::symbols("~/myproject")
syms$defs  # data.frame: name, file, line, exported
syms$calls # data.frame: caller, callee, file, line
```

Find who calls a function (and where the damage lands if you change it):

```r
saber::blast_radius("my_function", project = "~/myproject")
#>   caller      project      file         line
#>   do_thing    myproject    main.R         42
#>   run_batch   downstream   pipeline.R     17
```

Discover projects and their dependencies:

```r
saber::projects()
#>   package   title                  version  path            depends  imports
#>   saber     Context Engineering    0.7.0    /home/troy/saber        ...

saber::find_downstream("jsonlite")
#>   [1] "chatterbox" "cornfab" "diffuseR" "llamaR" "llm.api"
#>   [6] "safetensors" "stt.api" "torch" "tts.api" "tuber" "whisper"
```

Inspect any installed package:

```r
saber::pkg_exports("saber")
saber::pkg_help("symbols", "saber")
```

Render a call graph:

```r
svg <- saber::fn_graph("~/myproject")
writeLines(svg, "~/callgraph.svg")
```

## How it works

`agent_context()` loads standard context files for AI coding agents: project instructions (AGENTS.md / CLAUDE.md), Claude Code memory files, global instructions, and agent identity files (SOUL.md). It skips files the agent already autoloads to avoid duplication.

`briefing()` assembles project context from DESCRIPTION metadata, downstream dependents, and recent git commits. It writes the markdown to the user cache directory so both the agent and user see the same context.

`symbols()` runs `getParseData()` on every `R/*.R` file in a project, extracts function definitions and call sites, and caches the results as RDS. Cache invalidates on file content changes (MD5).

`blast_radius()` builds on `symbols()`. It finds internal callers, then scans `~/` for any project whose DESCRIPTION declares a dependency on the target package. Traces the call graph across all of them. With `include = c("r", "examples", "vignettes")` it also flags references in roxygen `@examples` blocks and vignette code chunks.

`fn_graph()` and `pkg_graph()` render force-directed SVG graphs via a base R Fruchterman-Reingold simulation. No JavaScript — tooltips and links work via native SVG features.

## Codex integration

Codex reads `AGENTS.md` files automatically before it starts work. This repo ships one at the root; for your own R projects, add rules like these so Codex reaches for saber instead of guessing:

```markdown
## saber Toolchain Rules

Before working on R code, use the right tool for the job:

| Situation | Command |
|-----------|---------|
| Understand a package's API | `Rscript -e 'saber::pkg_exports("pkg")'` |
| Read function docs | `Rscript -e 'saber::pkg_help("fn", "pkg")'` |
| Before renaming/changing a function | `Rscript -e 'saber::blast_radius("fn", project = ".")'` |
| Understand a project's call graph | `Rscript -e 'str(saber::symbols("."))'` |
| Discover R packages and deps | `Rscript -e 'saber::projects()'` |
| What depends on a package | `Rscript -e 'saber::find_downstream("pkg")'` |
| Project briefing | `Rscript -e 'saber::briefing("project")'` |

**blast_radius is mandatory before renaming, moving, or changing the signature of any exported function.** It finds every caller across this project and all downstream projects. Skip it and you break things silently.
```

### SessionStart hook

saber ships a hook script that injects a project briefing into Codex at the start of every session. Find it with:

```r
system.file("scripts", "session-start.R", package = "saber")
```

Enable hooks in your Codex config (`~/.codex/config.toml`):

```toml
[features]
hooks = true
```

You can also enable the same feature from the CLI:

```bash
codex --enable hooks
```

Then add the hook to `~/.codex/hooks.json`:

```json
{
  "hooks": {
    "SessionStart": [
      {
        "matcher": "startup|resume",
        "hooks": [
          {
            "type": "command",
            "command": "Rscript /path/to/session-start.R codex",
            "timeout": 15,
            "statusMessage": "Loading saber briefing"
          }
        ]
      }
    ]
  }
}
```

Codex may require new or changed hooks to be reviewed before they run. Open
`/hooks` in Codex and approve the `session-start.R` command after adding it.

If you want neutral cross-agent preferences injected too, create
`~/.config/agents/GLOBAL.md`. The hook appends it automatically after the
project briefing. Set `AGENTS_GLOBAL_MD` if you want a different path.

Every new Codex session starts with the project's metadata, downstream dependents, Claude Code memory (if available), recent git commits, and optional global preferences already in context.

## Claude Code integration

Add the following to your `~/.claude/CLAUDE.md` to teach Claude Code how to use saber:

```markdown
### saber Toolchain Rules

Before working on R code, use the right tool for the job:

| Situation | Command |
|-----------|---------|
| Understand a package's API | `Rscript -e 'saber::pkg_exports("pkg")'` |
| Read function docs | `Rscript -e 'saber::pkg_help("fn", "pkg")'` |
| Before renaming/changing a function | `Rscript -e 'saber::blast_radius("fn", project = ".")'` |
| Understand a project's call graph | `Rscript -e 'str(saber::symbols("."))'` |
| Discover R packages and deps | `Rscript -e 'saber::projects()'` |
| What depends on a package | `Rscript -e 'saber::find_downstream("pkg")'` |
| Project briefing | `Rscript -e 'saber::briefing("project")'` |

**blast_radius is mandatory before renaming, moving, or changing the signature of any exported function.** It finds every caller across this project and all downstream projects. Skip it and you break things silently.
```

### SessionStart hook

saber ships a hook script that injects a project briefing into Claude Code's context at the start of every session. Find it with:

```r
system.file("scripts", "session-start.R", package = "saber")
```

Then add it to your Claude Code settings (`~/.claude/settings.json`):

```json
{
  "hooks": {
    "SessionStart": [
      {
        "hooks": [
          {
            "type": "command",
            "command": "Rscript /path/to/session-start.R claude",
            "timeout": 15
          }
        ]
      }
    ]
  }
}
```

Every new session starts with the project's metadata, downstream dependents, and recent git commits already in context. The `claude` agent flag tells `briefing()` to skip Claude Code memory (which Claude Code autoloads separately).

## License

Apache-2.0

#!/usr/bin/env Rscript
# saber - generate project context at session start
# For use as a Claude Code or Codex SessionStart hook
#
# Usage: Rscript session-start.R [agent]
#   agent: "claude", "codex", or omit for interactive default

cli_args <- commandArgs(trailingOnly = TRUE)
agent <- if (length(cli_args) > 0L) cli_args[[1L]] else NULL

session_cwd <- getwd()

global_preferences_path <- function() {
    custom <- Sys.getenv("AGENTS_GLOBAL_MD", unset = "")
    if (nchar(custom) > 0L) {
        return(path.expand(custom))
    }
    path.expand("~/.config/agents/GLOBAL.md")
}

load_global_preferences <- function() {
    path <- global_preferences_path()
    if (!file.exists(path)) {
        return(NULL)
    }

    lines <- tryCatch(readLines(path, warn = FALSE),
                      error = function(e) NULL)
    if (is.null(lines)) {
        return(NULL)
    }

    text <- paste(lines, collapse = "\n")
    if (nchar(trimws(text)) == 0L) {
        return(NULL)
    }

    paste0("## Global Preferences\n\n", text, "\n")
}

resolve_repo_root <- function(path) {
    root <- tryCatch(
                    system2("git", c("-C", path, "rev-parse", "--show-toplevel"),
                            stdout = TRUE, stderr = FALSE),
                    error = function(e) character(0L)
    )
    if (length(root) == 0L) {
        return(NULL)
    }

    root <- trimws(root[[1L]])
    if (nchar(root) == 0L) {
        return(NULL)
    }
    path.expand(root)
}

load_saber_fun <- function(name, repo_root = NULL) {
    if (!is.null(repo_root)) {
        local_fun <- tryCatch(
                            {
                                desc <- file.path(repo_root, "DESCRIPTION")
                                r_dir <- file.path(repo_root, "R")
                                if (!file.exists(desc) || !dir.exists(r_dir)) {
                                    stop("No local package source found")
                                }
                                # Only source from the local repo when it IS
                                # saber. Other packages may export a symbol
                                # with the same name but a different
                                # signature.
                                pkg <- tryCatch(
                                                read.dcf(desc, fields = "Package")[1L, "Package"],
                                                error = function(e) NA_character_
                                )
                                if (is.na(pkg) || pkg != "saber") {
                                    stop("Local package is not saber")
                                }

                                env <- new.env(parent = baseenv())
                                r_files <- list.files(r_dir, pattern = "\\.[Rr]$",
                                                      full.names = TRUE)
                                for (f in r_files) {
                                    sys.source(f, envir = env)
                                }
                                get(name, envir = env, inherits = FALSE)
                            },
                            error = function(e) NULL
        )
        if (is.function(local_fun)) {
            return(local_fun)
        }
    }

    if (requireNamespace("saber", quietly = TRUE)) {
        return(getExportedValue("saber", name))
    }

    stop("saber not available")
}

load_agent_memory <- function(agent, project_dir, repo_root = NULL) {
    context_fun <- load_saber_fun("agent_context", repo_root)
    context_fun(agent = agent, project_dir = project_dir,
                include_project = FALSE,
                include_global = FALSE,
                include_soul = FALSE)
}

append_context <- function(text, section) {
    if (is.null(section) || nchar(trimws(section)) == 0L) {
        return(text)
    }
    paste0(text, "\n\n", section)
}

repo_root <- resolve_repo_root(session_cwd)
if (!is.null(repo_root)) {
    project <- basename(repo_root)
    scan_dir <- dirname(repo_root)
    project_dir <- repo_root
} else {
    project <- basename(session_cwd)
    scan_dir <- path.expand("~")
    project_dir <- session_cwd
}

briefing_text <- tryCatch(
    {
        briefing_fun <- load_saber_fun("briefing", repo_root)
        msg_lines <- utils::capture.output(
            briefing_fun(project, scan_dir = scan_dir),
            type = "message"
        )
        paste(msg_lines, collapse = "\n")
    },
    error = function(e) {
        paste0("# Briefing: ", project,
               "\n_saber not available:_ ", conditionMessage(e), "\n")
    }
)

if (is.null(briefing_text) || nchar(briefing_text) == 0L) {
    briefing_text <- paste0("# Briefing: ", project,
                            "\n_No briefing available._\n")
}

memory_text <- tryCatch(
    load_agent_memory(agent, project_dir, repo_root),
    error = function(e) NULL
)
briefing_text <- append_context(briefing_text, memory_text)

global_preferences <- load_global_preferences()
if (!is.null(global_preferences)) {
    briefing_text <- append_context(briefing_text, global_preferences)
}

escaped <- gsub("\\\\", "\\\\\\\\", briefing_text)
escaped <- gsub("\"", "\\\\\"", escaped)
escaped <- gsub("\n", "\\\\n", escaped)
escaped <- gsub("\t", "\\\\t", escaped)

cat(sprintf('{
  "hookSpecificOutput": {
    "hookEventName": "SessionStart",
    "additionalContext": "%s"
  }
}', escaped))

#' @title Documentation scanning
#' @description Scan roxygen \verb{@examples} blocks and vignette code for
#'   function references.

#' Scan roxygen \verb{@examples} blocks for function calls
#'
#' Reads each \code{R/*.R} file in a project, extracts the contents of every
#' roxygen \verb{@examples} (or \verb{@examplesIf}) block, and returns rows
#' for lines that contain a call to \code{fn}. Detection is regex based:
#' \code{fn} must appear followed by an opening parenthesis and must not be
#' part of a longer identifier. \verb{pkg::fn(...)} calls match as well.
#'
#' The \code{caller} column is the name of the function being documented
#' (the first \code{name <- function(...)} line after the roxygen block), or
#' the empty string if none is detected.
#'
#' @param project_dir Path to the project root.
#' @param fn Function name to search for.
#' @return A data.frame with columns caller, project, file, line, source.
#' @noRd
scan_examples <- function(project_dir, fn) {
    project_name <- basename(normalizePath(project_dir, mustWork = FALSE))
    empty <- empty_doc_results()

    r_dir <- file.path(project_dir, "R")
    if (!dir.exists(r_dir)) {
        return(empty)
    }

    r_files <- list.files(r_dir, pattern = "\\.[Rr]$", full.names = TRUE)
    if (length(r_files) == 0L) {
        return(empty)
    }

    results <- empty
    for (fp in r_files) {
        lines <- readLines(fp, warn = FALSE)
        blocks <- extract_example_blocks(lines)
        for (b in blocks) {
            hits <- match_fn_lines(lines[b$line_nums], fn)
            if (length(hits) == 0L) {
                next
            }
            results <- rbind(results,
                             data.frame(caller = b$documented_fn, project = project_name,
                                        file = basename(fp), line = b$line_nums[hits],
                                        source = "example", stringsAsFactors = FALSE))
        }
    }

    results
}

#' Scan vignette code chunks for function calls
#'
#' Looks in \code{vignettes/} (and \code{inst/doc/}) for Rmd, qmd, and Rnw
#' files. Extracts R code chunks and flags lines that call \code{fn}.
#'
#' @param project_dir Path to the project root.
#' @param fn Function name to search for.
#' @return A data.frame with columns caller, project, file, line, source.
#' @noRd
scan_vignettes <- function(project_dir, fn) {
    project_name <- basename(normalizePath(project_dir, mustWork = FALSE))
    empty <- empty_doc_results()

    dirs <- file.path(project_dir, c("vignettes", "inst/doc"))
    dirs <- dirs[dir.exists(dirs)]
    if (length(dirs) == 0L) {
        return(empty)
    }

    files <- unlist(lapply(dirs, list.files,
                           pattern = "\\.(Rmd|rmd|qmd|Rnw|rnw)$",
                           full.names = TRUE, recursive = TRUE))
    if (length(files) == 0L) {
        return(empty)
    }

    results <- empty
    for (fp in files) {
        ext <- tolower(tools::file_ext(fp))
        lines <- readLines(fp, warn = FALSE)
        is_sweave <- ext %in% c("rnw")
        in_chunk <- chunk_mask(lines, is_sweave)

        code_line_nums <- which(in_chunk)
        if (length(code_line_nums) == 0L) {
            next
        }
        hits <- match_fn_lines(lines[code_line_nums], fn)
        if (length(hits) == 0L) {
            next
        }
        # Strip project_dir prefix to get a relative path. Avoid regex
        # since Windows path separators (\) would be interpreted as
        # backreferences.
        rel <- fp
        for (sep in c("/", "\\")) {
            pfx <- paste0(project_dir, sep)
            if (startsWith(rel, pfx)) {
                rel <- substr(rel, nchar(pfx) + 1L, nchar(rel))
                break
            }
        }
        results <- rbind(results,
                         data.frame(caller = "<vignette>",
                                    project = project_name,
                                    file = rel,
                                    line = code_line_nums[hits],
                                    source = "vignette",
                                    stringsAsFactors = FALSE))
    }

    results
}

#' Extract roxygen \verb{@examples} blocks from a file's lines
#'
#' Returns a list of blocks, each a list with components \code{line_nums}
#' (1-indexed file line numbers of the example body lines) and
#' \code{documented_fn} (the name of the function defined after the block,
#' or the empty string).
#'
#' @noRd
extract_example_blocks <- function(lines) {
    n <- length(lines)
    blocks <- list()
    i <- 1L
    tag_re <- "^#'\\s*@\\w+"
    ex_re <- "^#'\\s*@examples(If)?\\b"

    while (i <= n) {
        if (grepl(ex_re, lines[i])) {
            line_nums <- integer()
            # If @examplesIf has content after the condition, the body starts
            # on the next line anyway. Ignore the tag line itself.
            i <- i + 1L
            while (i <= n && grepl("^#'", lines[i])) {
                if (grepl(tag_re, lines[i])) {
                    break
                }
                line_nums <- c(line_nums, i)
                i <- i + 1L
            }

            documented_fn <- next_defined_fn(lines, i, n)
            blocks[[length(blocks) + 1L]] <- list(line_nums = line_nums,
                documented_fn = documented_fn)
        } else {
            i <- i + 1L
        }
    }

    blocks
}

#' Find the next top-level function definition after a given line
#'
#' Scans forward from \code{start} looking for a \code{name <- function}
#' (or \code{name = function}) assignment at top level. Returns the name or
#' the empty string if not found within a few lines of non-roxygen code.
#'
#' @noRd
next_defined_fn <- function(lines, start, n) {
    pattern <- "^\\s*([A-Za-z.][A-Za-z0-9._]*)\\s*(<-|=)\\s*function\\b"
    j <- start
    while (j <= n) {
        # Skip blank lines and roxygen lines (in case multiple blocks abut)
        if (!nzchar(trimws(lines[j])) || grepl("^#'", lines[j])) {
            j <- j + 1L
            next
        }
        m <- regmatches(lines[j], regexec(pattern, lines[j]))
        if (length(m[[1L]]) >= 2L) {
            return(m[[1L]][2L])
        }
        # First non-blank non-roxygen line wasn't a function def; give up
        return("")
    }
    ""
}

#' Build a per-line logical mask for "this line is inside an R code chunk"
#'
#' @param lines Character vector of file lines.
#' @param is_sweave If TRUE, parse Rnw \verb{<<>>=...@} chunks; otherwise
#'   Rmd/qmd \verb{```{r...}...```} chunks.
#' @return Logical vector of \code{length(lines)}.
#' @noRd
chunk_mask <- function(lines, is_sweave = FALSE) {
    n <- length(lines)
    inside <- logical(n)
    in_chunk <- FALSE

    if (is_sweave) {
        start_re <- "^<<.*>>="
        end_re <- "^@\\s*$"
    } else {
        start_re <- "^```\\{[rR]([, }]|$)"
        end_re <- "^```\\s*$"
    }

    for (i in seq_len(n)) {
        if (!in_chunk) {
            if (grepl(start_re, lines[i])) {
                in_chunk <- TRUE
            }
            next
        }
        # in_chunk is TRUE
        if (grepl(end_re, lines[i])) {
            in_chunk <- FALSE
            next
        }
        inside[i] <- TRUE
    }

    inside
}

#' Find 1-indexed positions within \code{lines} containing a call to \code{fn}
#'
#' Matches \code{fn} followed by optional whitespace and an opening paren,
#' with a negative lookbehind that rejects identifier characters (so
#' \code{myfn(} or \code{my.fn(} do not match, but \code{pkg::fn(} does).
#'
#' @noRd
match_fn_lines <- function(lines, fn) {
    if (length(lines) == 0L) {
        return(integer())
    }
    pattern <- paste0("(?<![A-Za-z0-9._])",
                      gsub("([.\\\\+*?\\[\\]^$(){}|])", "\\\\\\1", fn, perl = TRUE),
                      "\\s*\\(")
    which(grepl(pattern, lines, perl = TRUE))
}

#' Empty doc-scan result frame
#' @noRd
empty_doc_results <- function() {
    data.frame(caller = character(), project = character(),
               file = character(), line = integer(), source = character(),
               stringsAsFactors = FALSE)
}

#' saber: Context Engineering for Large Language Model Agents
#'
#' Context-engineering primitives for Artificial Intelligence (AI) coding
#' agents working in R.
#' Assembles agent context from memory and instruction files, traces
#' function call blast radius across projects, generates project
#' briefings, parses R source into structured Abstract Syntax Tree
#' symbol indices, discovers dependency graphs, and introspects
#' installed packages.
#'
#' @keywords internal
"_PACKAGE"

#' @title Project briefings
#' @description Generate project context for AI coding agents.

#' Generate a project briefing
#'
#' Produces a concise markdown briefing combining DESCRIPTION metadata,
#' downstream dependents, and recent git commits. Written to the user
#' cache directory so both the agent and user see the same context.
#'
#' For runtime context (memory, identity files, project instructions),
#' see \code{\link{agent_context}}.
#'
#' @param project Project name. If NULL, inferred from the current working
#'   directory basename.
#' @param scan_dir Directory to scan for project directories.
#' @param briefs_dir Directory to write briefing markdown files.
#' @return The briefing text (character string), returned invisibly. Emitted
#'   via \code{message()} and written to \code{briefs_dir/{project}.md}.
#' @examples
#' d <- file.path(tempdir(), "briefpkg")
#' dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
#' writeLines(c("Package: briefpkg", "Title: Demo", "Version: 0.1.0"),
#'            file.path(d, "DESCRIPTION"))
#' briefing("briefpkg", scan_dir = tempdir(),
#'          briefs_dir = file.path(tempdir(), "briefs"))
#' @export
briefing <- function(project = NULL, scan_dir = path.expand("~"),
                     briefs_dir = file.path(tools::R_user_dir("saber", "cache"), "briefs")) {
    if (is.null(project)) {
        project <- basename(getwd())
    }
    dir.create(briefs_dir, recursive = TRUE, showWarnings = FALSE)

    lines <- character(0L)
    lines <- c(lines, sprintf("# Briefing: %s", project))
    lines <- c(lines,
               sprintf("_Generated %s_", format(Sys.time(), "%Y-%m-%d %H:%M")))
    lines <- c(lines, "")

    desc <- briefing_desc(project, scan_dir)
    if (length(desc) > 0L) {
        lines <- c(lines, desc, "")
    }

    ds <- briefing_downstream(project, scan_dir)
    if (length(ds) > 0L) {
        lines <- c(lines, ds, "")
    }

    git <- briefing_git(project, scan_dir)
    if (length(git) > 0L) {
        lines <- c(lines, git, "")
    }

    text <- paste(lines, collapse = "\n")

    outfile <- file.path(briefs_dir, paste0(project, ".md"))
    writeLines(lines, outfile)

    message(text)
    invisible(text)
}

#' DESCRIPTION metadata section
#' @noRd
briefing_desc <- function(project, scan_dir) {
    repo_dir <- file.path(scan_dir, project)
    desc_file <- file.path(repo_dir, "DESCRIPTION")
    if (!file.exists(desc_file)) {
        return(character(0L))
    }

    dcf <- tryCatch(
                    read.dcf(desc_file,
                             fields = c("Package", "Title", "Version", "Imports")),
                    error = function(e) NULL
    )
    if (is.null(dcf) || nrow(dcf) == 0L) {
        return(character(0L))
    }

    lines <- "## Package"
    pkg <- dcf[1L, "Package"]
    if (!is.na(pkg)) {
        lines <- c(lines, sprintf("- **Name**: %s", pkg))
    }

    title <- dcf[1L, "Title"]
    if (!is.na(title)) {
        lines <- c(lines, sprintf("- **Title**: %s", title))
    }

    ver <- dcf[1L, "Version"]
    if (!is.na(ver)) {
        lines <- c(lines, sprintf("- **Version**: %s", ver))
    }

    imports <- dcf[1L, "Imports"]
    if (!is.na(imports) && nchar(trimws(imports)) > 0L) {
        lines <- c(lines, sprintf("- **Imports**: %s", trimws(imports)))
    }

    if (length(lines) == 1L) {
        return(character(0L))
    }
    lines
}

#' Downstream dependents section
#' @noRd
briefing_downstream <- function(project, scan_dir) {
    ds <- find_downstream(project, scan_dir)
    if (length(ds) == 0L) {
        return(character(0L))
    }

    lines <- "## Downstream dependents"
    for (d in ds) {
        lines <- c(lines, sprintf("- %s", d))
    }
    lines
}

#' Recent git activity section
#' @noRd
briefing_git <- function(project, scan_dir) {
    repo_dir <- file.path(scan_dir, project)
    if (!dir.exists(file.path(repo_dir, ".git"))) {
        return(character(0L))
    }

    log <- tryCatch(
                    system2("git", c("-C", repo_dir, "log", "--oneline", "-5"),
                            stdout = TRUE, stderr = FALSE),
                    error = function(e) character(0L)
    )
    if (length(log) == 0L) {
        return(character(0L))
    }

    lines <- "## Recent commits"
    for (l in log) {
        lines <- c(lines, sprintf("- %s", l))
    }
    lines
}

# Tests for briefing()

# --- Setup ---
scan_dir <- file.path(tempdir(), "test_briefing")
briefs_dir <- file.path(tempdir(), "test_briefs")
dir.create(scan_dir, showWarnings = FALSE)

# Create a project with DESCRIPTION
pkg_dir <- file.path(scan_dir, "demopkg")
dir.create(pkg_dir, showWarnings = FALSE)
writeLines(c(
    "Package: demopkg",
    "Title: Demo Package",
    "Version: 0.1.0",
    "Imports: stats"
), file.path(pkg_dir, "DESCRIPTION"))

# Create a git repo with a commit
system2("git", c("-C", pkg_dir, "init", "-q"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.email", "test@test.com"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.name", "Test"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "add", "-A"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "commit", "-q", "-m", "init"),
        stdout = FALSE, stderr = FALSE)

# --- briefing() returns invisible character, emits via message() ---
msgs <- capture.output(
    result <- briefing("demopkg", scan_dir = scan_dir, briefs_dir = briefs_dir),
    type = "message"
)
expect_true(is.character(result))
expect_true(grepl("Briefing: demopkg", result))
expect_true(any(grepl("Briefing: demopkg", msgs)))

# --- briefing includes DESCRIPTION metadata ---
expect_true(grepl("Demo Package", result))
expect_true(grepl("0.1.0", result))

# --- briefing includes git log ---
expect_true(grepl("Recent commits", result))
expect_true(grepl("init", result))

# --- briefing writes to file ---
outfile <- file.path(briefs_dir, "demopkg.md")
expect_true(file.exists(outfile))

# --- briefing no longer includes memory (moved to agent_context) ---
expect_false(grepl("## Memory", result))

# --- briefing handles missing project gracefully ---
result_missing <- briefing("nonexistent", scan_dir = scan_dir,
                           briefs_dir = briefs_dir)
expect_true(is.character(result_missing))
expect_true(grepl("Briefing: nonexistent", result_missing))

# --- briefing_git() does not leak a system2 warning on a non-repo (#33) ---
# A directory with an invalid .git (passes dir.exists() but `git log` fails
# with status 128) must return character(0) silently, not leak a warning.
fakegit <- file.path(scan_dir, "fakegit")
dir.create(file.path(fakegit, ".git"), recursive = TRUE, showWarnings = FALSE)
expect_silent(git_out <- saber:::briefing_git("fakegit", scan_dir))
expect_equal(git_out, character(0L))

# A plain non-git directory is likewise silent and empty.
plain <- file.path(scan_dir, "plaindir")
dir.create(plain, showWarnings = FALSE)
expect_silent(plain_out <- saber:::briefing_git("plaindir", scan_dir))
expect_equal(plain_out, character(0L))

# --- Cleanup ---
unlink(scan_dir, recursive = TRUE)
unlink(briefs_dir, recursive = TRUE)

# Tests for briefing()

# --- Setup ---
scan_dir <- file.path(tempdir(), "test_briefing")
briefs_dir <- file.path(tempdir(), "test_briefs")
dir.create(scan_dir, showWarnings = FALSE)

# Create a project with DESCRIPTION
pkg_dir <- file.path(scan_dir, "demopkg")
dir.create(pkg_dir, showWarnings = FALSE)
writeLines(c(
    "Package: demopkg",
    "Title: Demo Package",
    "Version: 0.1.0",
    "Imports: stats"
), file.path(pkg_dir, "DESCRIPTION"))

# Create a git repo with a commit
system2("git", c("-C", pkg_dir, "init", "-q"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.email", "test@test.com"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.name", "Test"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "add", "-A"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "commit", "-q", "-m", "init"),
        stdout = FALSE, stderr = FALSE)

# --- briefing() returns invisible character, emits via message() ---
msgs <- capture.output(
    result <- briefing("demopkg", scan_dir = scan_dir, briefs_dir = briefs_dir),
    type = "message"
)
expect_true(is.character(result))
expect_true(grepl("Briefing: demopkg", result))
expect_true(any(grepl("Briefing: demopkg", msgs)))

# --- briefing includes DESCRIPTION metadata ---
expect_true(grepl("Demo Package", result))
expect_true(grepl("0.1.0", result))

# --- briefing includes git log ---
expect_true(grepl("Recent commits", result))
expect_true(grepl("init", result))

# --- briefing writes to file ---
outfile <- file.path(briefs_dir, "demopkg.md")
expect_true(file.exists(outfile))

# --- briefing no longer includes memory (moved to agent_context) ---
expect_false(grepl("## Memory", result))

# --- briefing handles missing project gracefully ---
result_missing <- briefing("nonexistent", scan_dir = scan_dir,
                           briefs_dir = briefs_dir)
expect_true(is.character(result_missing))
expect_true(grepl("Briefing: nonexistent", result_missing))

# --- Cleanup ---
unlink(scan_dir, recursive = TRUE)
unlink(briefs_dir, recursive = TRUE)

#' @title Package introspection
#' @description Query installed R packages for exported functions, internal
#'   functions, and help documentation.

#' List exported functions of a package
#'
#' Returns a data.frame of exported functions with their argument signatures.
#'
#' @param package Character. Package name.
#' @param pattern Optional regex to filter function names.
#' @return A data.frame with columns: name, args.
#' @examples
#' pkg_exports("tools")
#' pkg_exports("tools", pattern = "^Rd")
#' @export
pkg_exports <- function(package, pattern = NULL) {
    ns <- tryCatch(
                   getNamespace(package),
                   error = function(e) {
        stop("Package '", package, "' not found. Is it installed?",
             call. = FALSE)
    }
    )

    exports <- getNamespaceExports(ns)

    # Filter to functions only
    export_fns <- Filter(function(nm) {
        obj <- tryCatch(get(nm, envir = ns), error = function(e) NULL)
        is.function(obj)
    }, exports)

    if (!is.null(pattern)) {
        export_fns <- grep(pattern, export_fns, value = TRUE)
    }

    if (length(export_fns) == 0L) {
        return(data.frame(name = character(), args = character(),
                          stringsAsFactors = FALSE))
    }

    args_list <- vapply(export_fns, function(nm) {
        fn <- get(nm, envir = ns)
        paste(names(formals(fn)), collapse = ", ")
    }, character(1))

    data.frame(name = sort(export_fns), args = args_list[order(export_fns)],
               stringsAsFactors = FALSE, row.names = NULL)
}

#' List internal (non-exported) functions of a package
#'
#' Returns functions defined in a package namespace but not exported.
#'
#' @param package Character. Package name.
#' @param pattern Optional regex to filter function names.
#' @return A data.frame with columns: name, args.
#' @examples
#' pkg_internals("tools", pattern = "^check")
#' @export
pkg_internals <- function(package, pattern = NULL) {
    ns <- tryCatch(
                   getNamespace(package),
                   error = function(e) {
        stop("Package '", package, "' not found. Is it installed?",
             call. = FALSE)
    }
    )

    exports <- getNamespaceExports(ns)
    all_names <- ls(ns, all.names = TRUE)
    internal_names <- setdiff(all_names, exports)

    internal_fns <- Filter(function(nm) {
        obj <- get(nm, envir = ns)
        is.function(obj)
    }, internal_names)

    if (!is.null(pattern)) {
        internal_fns <- grep(pattern, internal_fns, value = TRUE)
    }

    if (length(internal_fns) == 0L) {
        return(data.frame(name = character(), args = character(),
                          stringsAsFactors = FALSE))
    }

    args_list <- vapply(internal_fns, function(nm) {
        fn <- get(nm, envir = ns)
        paste(names(formals(fn)), collapse = ", ")
    }, character(1))

    data.frame(name = sort(internal_fns),
               args = args_list[order(internal_fns)],
               stringsAsFactors = FALSE, row.names = NULL)
}

#' Get help for a package topic as markdown
#'
#' Extracts help documentation and converts it to clean markdown.
#'
#' @param topic Character. The help topic name.
#' @param package Character. Package name.
#' @param format Character. Output format: \code{"md"} (default) for plain
#'   markdown, or \code{"hugo"} for Hugo-compatible markdown with YAML front
#'   matter.
#' @return Character string of markdown help text.
#' @examples
#' cat(pkg_help("md5sum", "tools"))
#' @export
pkg_help <- function(topic, package, format = c("md", "hugo")) {
    format <- match.arg(format)
    db <- tools::Rd_db(package)

    rd_name <- paste0(topic, ".Rd")
    if (!rd_name %in% names(db)) {
        # Try alias match
        for (nm in names(db)) {
            rd <- db[[nm]]
            aliases <- rd_get_aliases(rd)
            if (topic %in% aliases) {
                rd_name <- nm
                break
            }
        }
    }

    if (!rd_name %in% names(db)) {
        stop("Topic '", topic, "' not found in package '", package, "'.",
             call. = FALSE)
    }

    rd <- db[[rd_name]]
    if (format == "hugo") {
        rd2hugo(rd, topic, package)
    } else {
        rd2md(rd)
    }
}

# --- Internal Rd-to-markdown conversion ---

#' Convert Rd object to markdown
#' @noRd
rd2md <- function(rd) {
    sections <- list()

    for (element in rd) {
        tag <- attr(element, "Rd_tag")
        if (is.null(tag)) {
            next
        }

        switch(tag,
               "\\title" = {
            sections$title <- trimws(rd_to_text(element))
        },
               "\\description" = {
            sections$description <- rd_to_md(element)
        },
               "\\usage" = {
            sections$usage <- rd_verbatim(element)
        },
               "\\arguments" = {
            sections$arguments <- rd_args_to_md(element)
        },
               "\\value" = {
            sections$value <- rd_to_md(element)
        },
               "\\details" = {
            sections$details <- rd_to_md(element)
        },
               "\\examples" = {
            sections$examples <- rd_verbatim(element)
        },
               "\\seealso" = {
            sections$seealso <- rd_to_md(element)
        },
               "\\references" = {
            sections$references <- rd_to_md(element)
        },
               "\\author" = {
            sections$author <- rd_to_md(element)
        },
               "\\note" = {
            sections$note <- rd_to_md(element)
        },
               "\\section" = {
            sec_title <- rd_to_text(element[[1]])
            sec_content <- rd_to_md(element[-1])
            if (is.null(sections$custom_sections)) {
                sections$custom_sections <- list()
            }
            sections$custom_sections <- c(sections$custom_sections,
                list(list(title = sec_title, content = sec_content)))
        }
        )
    }

    lines <- character()
    if (!is.null(sections$title)) {
        lines <- c(lines, paste0("### ", sections$title), "")
    }
    if (!is.null(sections$description)) {
        lines <- c(lines, "#### Description", "", sections$description, "")
    }
    if (!is.null(sections$usage)) {
        lines <- c(lines, "#### Usage", "", "```r", sections$usage, "```", "")
    }
    if (!is.null(sections$arguments)) {
        lines <- c(lines, "#### Arguments", "", sections$arguments, "")
    }
    if (!is.null(sections$details)) {
        lines <- c(lines, "#### Details", "", sections$details, "")
    }
    if (!is.null(sections$value)) {
        lines <- c(lines, "#### Value", "", sections$value, "")
    }
    if (!is.null(sections$custom_sections)) {
        for (sec in sections$custom_sections) {
            lines <- c(lines, paste0("#### ", sec$title), "", sec$content, "")
        }
    }
    if (!is.null(sections$note)) {
        lines <- c(lines, "#### Note", "", sections$note, "")
    }
    if (!is.null(sections$seealso)) {
        lines <- c(lines, "#### See Also", "", sections$seealso, "")
    }
    if (!is.null(sections$references)) {
        lines <- c(lines, "#### References", "", sections$references, "")
    }
    if (!is.null(sections$author)) {
        lines <- c(lines, "#### Author", "", sections$author, "")
    }
    if (!is.null(sections$examples)) {
        lines <- c(lines, "#### Examples", "", "```r", sections$examples,
                   "```", "")
    }

    paste(lines, collapse = "\n")
}

#' Convert Rd object to Hugo-compatible markdown
#' @noRd
rd2hugo <- function(rd, topic, package) {
    title <- ""
    description <- ""
    for (element in rd) {
        tag <- attr(element, "Rd_tag")
        if (identical(tag, "\\title")) {
            title <- trimws(rd_to_text(element))
        }
        if (identical(tag, "\\description")) {
            description <- trimws(rd_to_md(element))
        }
    }

    body <- rd2md(rd)

    front <- paste0("---\n", "title: \"", topic, "\"\n", "package: \"",
                    package, "\"\n", "description: >-\n", "  ",
                    gsub("\n", " ", description), "\n", "---\n")

    paste0(front, "\n", body)
}

#' Extract plain text from Rd element
#' @noRd
rd_to_text <- function(element) {
    if (is.character(element)) {
        return(element)
    }
    paste(unlist(lapply(element, rd_to_text)), collapse = "")
}

#' Convert Rd content to markdown with tag handling
#' @noRd
rd_to_md <- function(element) {
    if (is.character(element)) {
        return(element)
    }

    result <- character()
    for (child in element) {
        tag <- attr(child, "Rd_tag")
        if (is.null(tag)) {
            if (is.character(child)) {
                result <- c(result, child)
            } else {
                result <- c(result, rd_to_md(child))
            }
        } else {
            content <- rd_to_text(child)
            md <- switch(tag,
                         "\\code" = paste0("`", content, "`"),
                         "\\link" = paste0("`", content, "`"),
                         "\\linkS4class" = paste0("`", content, "`"),
                         "\\pkg" = paste0("**", content, "**"),
                         "\\emph" = paste0("*", content, "*"),
                         "\\strong" = paste0("**", content, "**"),
                         "\\bold" = paste0("**", content, "**"),
                         "\\sQuote" = paste0("'", content, "'"),
                         "\\dQuote" = paste0("\"", content, "\""),
                         "\\file" = paste0("`", content, "`"),
                         "\\url" = content,
                         "\\href" = {
                if (length(child) >= 2) {
                    url <- rd_to_text(child[[1]])
                    text <- rd_to_text(child[[2]])
                    paste0("[", text, "](", url, ")")
                } else {
                    content
                }
            },
                         "\\email" = content,
                         "\\var" = paste0("*", content, "*"),
                         "\\env" = paste0("`", content, "`"),
                         "\\option" = paste0("`", content, "`"),
                         "\\command" = paste0("`", content, "`"),
                         "\\dfn" = paste0("*", content, "*"),
                         "\\acronym" = content,
                         "\\dots" = "...",
                         "\\ldots" = "...",
                         "\\cr" = "\n",
                         "\\tab" = "\t",
                         "\\R" = "R",
                         "\\describe" = rd_describe_to_md(child),
                         "\\itemize" = rd_itemize_to_md(child),
                         "\\enumerate" = rd_enumerate_to_md(child),
                         "\\item" = rd_to_md(child),
                         "RCODE" = content,
                         "TEXT" = content,
                         "VERB" = content,
                         "COMMENT" = "",
                         rd_to_md(child)
            )
            result <- c(result, md)
        }
    }

    text <- paste(result, collapse = "")
    text <- rd_postprocess(text)
    trimws(text)
}

#' Extract verbatim text (usage, examples)
#' @noRd
rd_verbatim <- function(element) {
    text <- rd_to_text(element)
    gsub("^\\n+|\\n+$", "", text)
}

#' Convert \\arguments to markdown
#' @noRd
rd_args_to_md <- function(element) {
    lines <- character()
    for (child in element) {
        tag <- attr(child, "Rd_tag")
        if (identical(tag, "\\item") && length(child) >= 2) {
            arg_name <- rd_to_text(child[[1]])
            arg_desc <- rd_to_md(child[[2]])
            lines <- c(lines, paste0("- **`", arg_name, "`**: ", arg_desc))
        }
    }
    paste(lines, collapse = "\n")
}

#' Convert \\describe to markdown
#' @noRd
rd_describe_to_md <- function(element) {
    lines <- character()
    for (child in element) {
        tag <- attr(child, "Rd_tag")
        if (identical(tag, "\\item") && length(child) >= 2) {
            item_name <- rd_to_text(child[[1]])
            item_desc <- rd_to_md(child[[2]])
            lines <- c(lines, paste0("- **", item_name, "**: ", item_desc))
        }
    }
    paste(lines, collapse = "\n")
}

#' Convert \\itemize to markdown
#' @noRd
rd_itemize_to_md <- function(element) {
    lines <- character()
    for (child in element) {
        tag <- attr(child, "Rd_tag")
        if (identical(tag, "\\item")) {
            lines <- c(lines, paste0("- ", rd_to_md(child)))
        }
    }
    paste(lines, collapse = "\n")
}

#' Convert \\enumerate to markdown
#' @noRd
rd_enumerate_to_md <- function(element) {
    lines <- character()
    i <- 1L
    for (child in element) {
        tag <- attr(child, "Rd_tag")
        if (identical(tag, "\\item")) {
            lines <- c(lines, paste0(i, ". ", rd_to_md(child)))
            i <- i + 1L
        }
    }
    paste(lines, collapse = "\n")
}

#' Post-process text with embedded Rd markup
#' @noRd
rd_postprocess <- function(text) {
    text <- gsub("\\\\describe\\{\\s*\n?", "", text)
    text <- gsub("\\s*\\\\item\\{([^}]+)\\}\\{([^}]+)\\}", "\n- **\\1**: \\2",
                 text)
    text <- gsub("\n\\s*\\}\n", "\n", text)
    text <- gsub("^\\s*\\}\\s*$", "", text)
    text <- gsub("\\\\code\\{([^}]+)\\}", "`\\1`", text)
    text <- gsub("\\\\link\\{([^}]+)\\}", "`\\1`", text)
    text <- gsub("\\\\emph\\{([^}]+)\\}", "*\\1*", text)
    text <- gsub("\\\\strong\\{([^}]+)\\}", "**\\1**", text)
    text <- gsub("\\\\bold\\{([^}]+)\\}", "**\\1**", text)
    text <- gsub("\\\\pkg\\{([^}]+)\\}", "**\\1**", text)
    text <- gsub("\\\\file\\{([^}]+)\\}", "`\\1`", text)
    text <- gsub("\\\\sQuote\\{([^}]+)\\}", "'\\1'", text)
    text <- gsub("\\\\dQuote\\{([^}]+)\\}", "\"\\1\"", text)
    text <- gsub("\\\\dots\\b", "...", text)
    text <- gsub("\\\\ldots\\b", "...", text)
    text <- gsub("\\\\R\\b", "R", text)
    text <- gsub("\n{3,}", "\n\n", text)
    trimws(text)
}

#' Get aliases from Rd object
#' @noRd
rd_get_aliases <- function(rd) {
    aliases <- character()
    for (tag in rd) {
        if (identical(attr(tag, "Rd_tag"), "\\alias")) {
            aliases <- c(aliases, as.character(tag))
        }
    }
    trimws(aliases)
}

# Tests for pkg.R

library(saber)

# --- pkg_exports ---

# saber itself is a good test subject
exp <- pkg_exports("saber")
expect_true(is.data.frame(exp))
expect_true(nrow(exp) > 0L)
expect_true("name" %in% names(exp))
expect_true("args" %in% names(exp))
expect_true("symbols" %in% exp$name)
expect_true("blast_radius" %in% exp$name)

# Pattern filter
exp2 <- pkg_exports("saber", pattern = "^pkg_")
expect_true(nrow(exp2) >= 2L)
expect_true(all(grepl("^pkg_", exp2$name)))

# Non-existent package
expect_error(pkg_exports("nonexistent_pkg_12345"))

# --- pkg_internals ---

int <- pkg_internals("saber")
expect_true(is.data.frame(int))
expect_true(nrow(int) > 0L)
# file_hash is internal
expect_true("file_hash" %in% int$name)

# Pattern filter
int2 <- pkg_internals("saber", pattern = "^file_")
expect_true(nrow(int2) >= 1L)
expect_true(all(grepl("^file_", int2$name)))

# --- pkg_help ---

# Get help for a known topic
md <- pkg_help("symbols", "saber")
expect_true(is.character(md))
expect_true(nchar(md) > 0L)
expect_true(grepl("project_dir|symbol|AST", md))

# Non-existent topic
expect_error(pkg_help("nonexistent_topic_xyz", "saber"))

#' @title Code intelligence: AST symbol index
#' @description Parse R source files into structured function definitions and
#'   call relationships.
#' @importFrom utils getParseData

#' Build a symbol index for a project
#'
#' Parses all \code{R/*.R} files in a project directory using
#' \code{getParseData()} to extract function definitions and function calls.
#' Results are cached as RDS in the user cache directory.
#'
#' @param project_dir Path to the project directory.
#' @param cache_dir Directory for symbol cache files.
#' @return A list with components:
#'   \describe{
#'     \item{defs}{data.frame(name, file, line, exported)}
#'     \item{calls}{data.frame(caller, callee, file, line)}
#'   }
#' @examples
#' # Create a minimal project with R source files
#' d <- file.path(tempdir(), "demopkg")
#' dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
#' writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
#' writeLines("double <- function(x) add(x, x)", file.path(d, "R", "double.R"))
#'
#' idx <- symbols(d, cache_dir = tempdir())
#' idx$defs   # function definitions
#' idx$calls  # call relationships (double calls add)
#' @export
symbols <- function(project_dir,
                    cache_dir = file.path(tools::R_user_dir("saber", "cache"), "symbols")) {
    project_dir <- normalizePath(project_dir, mustWork = TRUE)
    project_name <- basename(project_dir)

    r_dir <- file.path(project_dir, "R")
    if (!dir.exists(r_dir)) {
        return(list(defs = data.frame(name = character(), file = character(),
                                      line = integer(), exported = logical(),
                                      stringsAsFactors = FALSE),
                    calls = data.frame(caller = character(), callee = character(),
                                       file = character(), line = integer(),
                                       stringsAsFactors = FALSE)))
    }

    r_files <- list.files(r_dir, pattern = "\\.[Rr]$", full.names = TRUE)
    if (length(r_files) == 0L) {
        return(list(defs = data.frame(name = character(), file = character(),
                                      line = integer(), exported = logical(),
                                      stringsAsFactors = FALSE),
                    calls = data.frame(caller = character(), callee = character(),
                                       file = character(), line = integer(),
                                       stringsAsFactors = FALSE)))
    }

    # Check cache
    dir.create(cache_dir, recursive = TRUE, showWarnings = FALSE)
    cache_file <- file.path(cache_dir, paste0(project_name, ".rds"))

    hashes <- vapply(r_files, file_hash, character(1))
    hash_key <- paste(sort(paste(basename(r_files), hashes)), collapse = "|")

    if (file.exists(cache_file)) {
        cached <- readRDS(cache_file)
        if (identical(cached$hash_key, hash_key)) {
            return(cached$result)
        }
    }

    # Get exported names from NAMESPACE if available
    ns_file <- file.path(project_dir, "NAMESPACE")
    exported_names <- character()
    if (file.exists(ns_file)) {
        ns_lines <- readLines(ns_file, warn = FALSE)
        # Match export(name) patterns
        m <- regmatches(ns_lines, regexec("^export\\(([^)]+)\\)", ns_lines))
        exported_names <- unlist(lapply(m, function(x) {
            if (length(x) > 1) x[2] else character()
        }))
        # Also match S3method(generic,class)
        m2 <- regmatches(ns_lines,
                         regexec("^S3method\\(([^,]+),([^)]+)\\)", ns_lines))
        s3_names <- unlist(lapply(m2, function(x) {
            if (length(x) > 1) paste0(x[2], ".", x[3]) else character()
        }))
        exported_names <- c(exported_names, s3_names)
    }

    all_defs <- data.frame(name = character(), file = character(),
                           line = integer(), exported = logical(),
                           stringsAsFactors = FALSE)
    all_calls <- data.frame(caller = character(), callee = character(),
                            file = character(), line = integer(),
                            stringsAsFactors = FALSE)

    for (fp in r_files) {
        rel_file <- basename(fp)
        parsed <- tryCatch({
            expr <- parse(fp, keep.source = TRUE)
            getParseData(expr, includeText = TRUE)
        }, error = function(e) NULL)

        if (is.null(parsed) || nrow(parsed) == 0L) {
            next
        }

        # Extract function definitions: SYMBOL followed by LEFT_ASSIGN/EQ_ASSIGN
        # then FUNCTION
        file_defs <- extract_defs(parsed, rel_file, exported_names)
        all_defs <- rbind(all_defs, file_defs)

        # Extract function calls
        file_calls <- extract_calls(parsed, rel_file, file_defs$name)
        all_calls <- rbind(all_calls, file_calls)
    }

    result <- list(defs = all_defs, calls = all_calls)

    # Cache
    saveRDS(list(hash_key = hash_key, result = result), cache_file)

    result
}

#' Extract function definitions from parse data
#' @noRd
extract_defs <- function(pd, file, exported_names) {
    # Strategy: find LEFT_ASSIGN/EQ_ASSIGN tokens, then look at siblings
    # of the parent expr. In R's parse tree:
    #   expr(parent) -> expr(lhs) -> SYMBOL, LEFT_ASSIGN, expr(rhs) -> FUNCTION
    assigns <- pd[pd$token %in% c("LEFT_ASSIGN", "EQ_ASSIGN"),, drop = FALSE]
    defs <- data.frame(name = character(), file = character(),
                       line = integer(), exported = logical(),
                       stringsAsFactors = FALSE)

    for (i in seq_len(nrow(assigns))) {
        a <- assigns[i,]
        parent_id <- a$parent
        siblings <- pd[pd$parent == parent_id,, drop = FALSE]
        siblings <- siblings[order(siblings$line1, siblings$col1),, drop = FALSE]

        assign_pos <- which(siblings$id == a$id)
        if (assign_pos < 2L || assign_pos >= nrow(siblings)) {
            next
        }

        # LHS is an expr wrapper; look inside it for SYMBOL
        lhs_expr <- siblings[assign_pos - 1L,]
        if (lhs_expr$token == "SYMBOL") {
            fn_name <- lhs_expr$text
            fn_line <- lhs_expr$line1
        } else if (lhs_expr$token == "expr") {
            lhs_children <- pd[pd$parent == lhs_expr$id,, drop = FALSE]
            sym <- lhs_children[lhs_children$token == "SYMBOL",, drop = FALSE]
            if (nrow(sym) == 0L) {
                next
            }
            fn_name <- sym$text[1]
            fn_line <- sym$line1[1]
        } else {
            next
        }

        # RHS is an expr wrapper; look inside for FUNCTION token
        rhs_expr <- siblings[assign_pos + 1L,]
        has_function <- FALSE
        if (rhs_expr$token == "expr") {
            rhs_children <- pd[pd$parent == rhs_expr$id,, drop = FALSE]
            has_function <- any(rhs_children$token == "FUNCTION")
        }

        if (has_function) {
            defs <- rbind(defs,
                          data.frame(name = fn_name, file = file, line = fn_line,
                                     exported = fn_name %in% exported_names,
                                     stringsAsFactors = FALSE))
        }
    }

    defs
}

#' Extract function calls from parse data
#' @noRd
extract_calls <- function(pd, file, local_defs) {
    # Build a line-range map: for each function def, find the extent of its
    # top-level expr (the enclosing assignment expr)
    def_ranges <- build_def_ranges(pd, local_defs)

    # SYMBOL_FUNCTION_CALL tokens are function calls
    call_tokens <- pd[pd$token == "SYMBOL_FUNCTION_CALL",, drop = FALSE]

    # Also find namespace calls: pkg::fn
    ns_calls <- pd[pd$token == "NS_GET",, drop = FALSE]

    calls <- data.frame(caller = character(), callee = character(),
                        file = character(), line = integer(),
                        stringsAsFactors = FALSE)

    for (i in seq_len(nrow(call_tokens))) {
        ct <- call_tokens[i,]
        caller <- enclosing_def(ct$line1, def_ranges)
        calls <- rbind(calls,
                       data.frame(caller = caller, callee = ct$text, file = file,
                                  line = ct$line1, stringsAsFactors = FALSE))
    }

    for (i in seq_len(nrow(ns_calls))) {
        nc <- ns_calls[i,]
        parent_id <- nc$parent
        siblings <- pd[pd$parent == parent_id,, drop = FALSE]
        siblings <- siblings[order(siblings$line1, siblings$col1),, drop = FALSE]

        pkg_token <- siblings[siblings$token == "SYMBOL_PACKAGE",, drop = FALSE]
        fn_token <- siblings[siblings$token %in% c("SYMBOL_FUNCTION_CALL",
                "SYMBOL"),, drop = FALSE]

        if (nrow(pkg_token) > 0L && nrow(fn_token) > 0L) {
            caller <- enclosing_def(nc$line1, def_ranges)
            callee <- paste0(pkg_token$text[1], "::", fn_token$text[1])
            calls <- rbind(calls,
                           data.frame(caller = caller, callee = callee, file = file,
                                      line = nc$line1, stringsAsFactors = FALSE))
        }
    }

    calls
}

#' Build line-range map for function definitions
#' @noRd
build_def_ranges <- function(pd, local_defs) {
    if (length(local_defs) == 0L) {
        return(NULL)
    }

    assigns <- pd[pd$token %in% c("LEFT_ASSIGN", "EQ_ASSIGN"),, drop = FALSE]
    ranges <- data.frame(name = character(), start = integer(),
                         end = integer(), stringsAsFactors = FALSE)

    for (i in seq_len(nrow(assigns))) {
        a <- assigns[i,]
        parent_id <- a$parent
        parent_row <- pd[pd$id == parent_id,, drop = FALSE]
        if (nrow(parent_row) == 0L) {
            next
        }

        siblings <- pd[pd$parent == parent_id,, drop = FALSE]
        siblings <- siblings[order(siblings$line1, siblings$col1),, drop = FALSE]
        assign_pos <- which(siblings$id == a$id)
        if (assign_pos < 2L) {
            next
        }

        lhs_expr <- siblings[assign_pos - 1L,]
        fn_name <- NULL
        if (lhs_expr$token == "SYMBOL" && lhs_expr$text %in% local_defs) {
            fn_name <- lhs_expr$text
        } else if (lhs_expr$token == "expr") {
            lhs_children <- pd[pd$parent == lhs_expr$id,, drop = FALSE]
            sym <- lhs_children[lhs_children$token == "SYMBOL",, drop = FALSE]
            if (nrow(sym) > 0L && sym$text[1] %in% local_defs) {
                fn_name <- sym$text[1]
            }
        }

        if (!is.null(fn_name)) {
            ranges <- rbind(ranges,
                            data.frame(name = fn_name, start = parent_row$line1,
                                       end = parent_row$line2, stringsAsFactors = FALSE))
        }
    }

    ranges
}

#' Find which function definition encloses a line
#' @noRd
enclosing_def <- function(line, def_ranges) {
    if (is.null(def_ranges) || nrow(def_ranges) == 0L) {
        return("<top-level>")
    }
    inside <- def_ranges[def_ranges$start <= line &
        def_ranges$end >= line,, drop = FALSE]
    if (nrow(inside) == 0L) {
        return("<top-level>")
    }
    # If nested, pick the innermost (smallest range)
    inside$span <- inside$end - inside$start
    inside$name[which.min(inside$span)]
}

# Tests for blast.R

library(saber)

# --- Setup: create a fake project with functions ---

fake_home <- tempfile("home")
proj <- file.path(fake_home, "mypkg")
dir.create(file.path(proj, "R"), recursive = TRUE)

writeLines(c(
  "helper <- function(x) x + 1",
  "",
  "main_fn <- function(x) {",
  "  helper(x)",
  "}"
), file.path(proj, "R", "code.R"))

writeLines("export(main_fn)", file.path(proj, "NAMESPACE"))

# Create a downstream project that imports mypkg
downstream <- file.path(fake_home, "downstream")
dir.create(file.path(downstream, "R"), recursive = TRUE)

writeLines(c(
  "Package: downstream",
  "Version: 0.1.0",
  "Title: Downstream",
  "Imports: mypkg"
), file.path(downstream, "DESCRIPTION"))

writeLines(c(
  "use_helper <- function() {",
  "  mypkg::helper(42)",
  "}"
), file.path(downstream, "R", "code.R"))

cache <- tempfile("symcache")

# --- blast_radius for an internal function ---

br <- blast_radius("helper", project = proj, scan_dir = fake_home,
                   cache_dir = cache)
expect_true(is.data.frame(br))
expect_true(nrow(br) >= 1L)
expect_true("main_fn" %in% br$caller)

# --- source column is populated with "r" for R source hits ---
expect_true("source" %in% names(br))
expect_true(all(br$source == "r"))

# --- blast_radius finds downstream callers ---

# The downstream project imports mypkg and calls mypkg::helper
expect_true(any(br$project == "downstream"))

# --- blast_radius for a non-existent function ---

br2 <- blast_radius("nonexistent_fn_xyz", project = proj,
                    scan_dir = fake_home, cache_dir = cache)
expect_equal(nrow(br2), 0L)
expect_true("source" %in% names(br2))

# --- include validation ---
expect_error(blast_radius("helper", project = proj, include = "bogus",
                          scan_dir = fake_home, cache_dir = cache))

# --- include = "examples": scans roxygen @examples blocks ---

# Add a second function whose roxygen @examples calls helper()
writeLines(c(
  "#' Documented fn",
  "#'",
  "#' @param x input",
  "#' @examples",
  "#' helper(1)",
  "#' \\dontrun{",
  "#' helper(2)",
  "#' }",
  "#' @export",
  "documented <- function(x) x",
  "",
  "#' Other fn with prose only",
  "#' @examples",
  "#' # helper is mentioned in a comment but not called",
  "#' 1 + 1",
  "#' @export",
  "other <- function(x) x"
), file.path(proj, "R", "documented.R"))

# Invalidate cache so symbols() re-reads
unlink(cache, recursive = TRUE)

br_ex <- blast_radius("helper", project = proj,
                      include = c("r", "examples"),
                      scan_dir = fake_home, cache_dir = cache)
ex_rows <- br_ex[br_ex$source == "example",, drop = FALSE]
expect_true(nrow(ex_rows) >= 2L)        # two calls inside @examples
expect_equal(unique(ex_rows$file), "documented.R")
expect_true("documented" %in% ex_rows$caller)
# The prose-only @examples block in `other` should NOT match
expect_false(any(ex_rows$caller == "other"))

# Without examples, the ex hits disappear
br_no_ex <- blast_radius("helper", project = proj, include = "r",
                         scan_dir = fake_home, cache_dir = cache)
expect_false(any(br_no_ex$source == "example"))

# --- include = "vignettes": scans Rmd code chunks ---

dir.create(file.path(proj, "vignettes"))
writeLines(c(
  "---",
  "title: demo",
  "---",
  "",
  "Some prose that mentions helper() but is not in a chunk.",
  "",
  "```{r setup}",
  "helper(10)",
  "```",
  "",
  "More prose.",
  "",
  "```{r, eval=FALSE}",
  "mypkg::helper(20)",
  "```"
), file.path(proj, "vignettes", "demo.Rmd"))

br_vi <- blast_radius("helper", project = proj,
                      include = c("r", "vignettes"),
                      scan_dir = fake_home, cache_dir = cache)
vi_rows <- br_vi[br_vi$source == "vignette",, drop = FALSE]
expect_equal(nrow(vi_rows), 2L)
expect_true(all(grepl("demo\\.Rmd$", vi_rows$file)))
# Prose mention outside chunks is not counted
expect_false(any(vi_rows$line == 5L))

# --- include accepts all three at once ---
br_all <- blast_radius("helper", project = proj,
                       include = c("r", "examples", "vignettes"),
                       scan_dir = fake_home, cache_dir = cache)
expect_true(all(c("r", "example", "vignette") %in% br_all$source))

# --- Cleanup ---
unlink(c(fake_home, cache), recursive = TRUE)

# Tests for symbols.R

library(saber)

# --- symbols on saber itself ---

cache <- tempfile("symcache")
syms <- symbols(system.file(package = "saber"), cache_dir = cache)

# Should return a list with defs and calls
expect_true(is.list(syms))
expect_true("defs" %in% names(syms))
expect_true("calls" %in% names(syms))
expect_true(is.data.frame(syms$defs))
expect_true(is.data.frame(syms$calls))

# --- symbols on a temp project ---

proj <- tempfile("proj")
dir.create(file.path(proj, "R"), recursive = TRUE)

writeLines(c(
  "my_add <- function(x, y) {",
  "  x + y",
  "}",
  "",
  "my_mul <- function(x, y) {",
  "  result <- my_add(x, 0)",
  "  x * y",
  "}"
), file.path(proj, "R", "math.R"))

writeLines("export(my_add)", file.path(proj, "NAMESPACE"))

syms2 <- symbols(proj, cache_dir = cache)

# Defs
expect_true(nrow(syms2$defs) >= 2L)
expect_true("my_add" %in% syms2$defs$name)
expect_true("my_mul" %in% syms2$defs$name)

# my_add should be exported (it's in NAMESPACE)
add_row <- syms2$defs[syms2$defs$name == "my_add", ]
expect_true(add_row$exported)

# my_mul should not be exported
mul_row <- syms2$defs[syms2$defs$name == "my_mul", ]
expect_false(mul_row$exported)

# Calls: my_mul calls my_add
add_calls <- syms2$calls[syms2$calls$callee == "my_add", ]
expect_true(nrow(add_calls) >= 1L)

# --- Cache works ---

# Second call should use cache (same hashes)
syms3 <- symbols(proj, cache_dir = cache)
expect_equal(syms3$defs$name, syms2$defs$name)

# --- Empty project ---

empty_proj <- tempfile("emptyproj")
dir.create(empty_proj)
empty_syms <- symbols(empty_proj, cache_dir = cache)
expect_equal(nrow(empty_syms$defs), 0L)
expect_equal(nrow(empty_syms$calls), 0L)

# --- Cleanup ---
unlink(c(cache, proj, empty_proj), recursive = TRUE)

#' @title Blast radius analysis
#' @description Find all callers of a function across projects.

#' Find callers of a function across projects
#'
#' Given a function name and project, finds all internal callers within that
#' project and all callers in downstream projects (projects whose DESCRIPTION
#' lists this one in Depends, Imports, or LinkingTo).
#'
#' With \code{include = c("r", "examples", "vignettes")} the search can be
#' extended to references in the target project's roxygen \verb{@examples}
#' blocks and vignette code chunks (Rmd, qmd, Rnw). Documentation scanning is
#' target-project only; it does not walk downstream projects' docs.
#'
#' @param fn Character. Function name to search for.
#' @param project Character. Project name (or path to project directory).
#' @param include Character vector. Any of \code{"r"} (R source, default),
#'   \code{"examples"} (roxygen \verb{@examples} blocks in the target
#'   project), and \code{"vignettes"} (code chunks in the target project's
#'   vignettes).
#' @param scan_dir Directory to scan for downstream projects.
#' @param cache_dir Directory for symbol cache files.
#' @param exclude Character vector of directory basenames to skip when
#'   scanning for downstream projects.
#' @return A data.frame with columns: caller, project, file, line, source.
#'   \code{source} is one of \code{"r"}, \code{"example"}, \code{"vignette"}.
#' @examples
#' # Create a minimal project
#' d <- file.path(tempdir(), "blastpkg")
#' dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
#' writeLines("helper <- function(x) x + 1", file.path(d, "R", "helper.R"))
#' writeLines("main <- function(x) helper(x * 2)", file.path(d, "R", "main.R"))
#'
#' # Find all callers of helper()
#' blast_radius("helper", project = d, scan_dir = tempdir(),
#'              cache_dir = tempdir())
#'
#' # Include roxygen @examples and vignettes from the target project
#' blast_radius("helper", project = d, include = c("r", "examples", "vignettes"),
#'              scan_dir = tempdir(), cache_dir = tempdir())
#' @export
blast_radius <- function(fn, project = NULL, include = "r",
                         scan_dir = path.expand("~"),
                         cache_dir = file.path(tools::R_user_dir("saber", "cache"), "symbols"),
                         exclude = default_exclude()) {
    allowed <- c("r", "examples", "vignettes")
    bad <- setdiff(include, allowed)
    if (length(bad) > 0L) {
        stop("invalid 'include' value(s): ", paste(bad, collapse = ", "),
             ". Allowed: ", paste(allowed, collapse = ", "))
    }

    if (is.null(project)) {
        project <- basename(getwd())
    }

    # Resolve project directory
    project_dir <- project
    if (!dir.exists(file.path(project, "R"))) {
        project_dir <- file.path(path.expand("~"), project)
    }
    project_name <- basename(normalizePath(project_dir, mustWork = FALSE))

    results <- empty_blast_results()

    # 1. Internal callers from this project's symbol cache
    if ("r" %in% include && dir.exists(project_dir)) {
        syms <- symbols(project_dir, cache_dir = cache_dir)
        internal <- syms$calls[syms$calls$callee == fn,, drop = FALSE]
        if (nrow(internal) > 0L) {
            results <- rbind(results,
                             data.frame(caller = internal$caller,
                                        project = project_name,
                                        file = internal$file,
                                        line = internal$line,
                                        source = "r",
                                        stringsAsFactors = FALSE))
        }
    }

    # 2. Downstream projects via DESCRIPTION files (R source only)
    if ("r" %in% include) {
        downstream <- find_downstream(project_name, scan_dir, exclude)
        for (ds_name in downstream) {
            ds_dir <- file.path(scan_dir, ds_name)
            if (!dir.exists(file.path(ds_dir, "R"))) {
                next
            }

            ds_syms <- symbols(ds_dir, cache_dir = cache_dir)
            qualified <- paste0(project_name, "::", fn)
            ds_callers <- ds_syms$calls[ds_syms$calls$callee == qualified |
                ds_syms$calls$callee == fn,, drop = FALSE]
            if (nrow(ds_callers) > 0L) {
                results <- rbind(results,
                                 data.frame(caller = ds_callers$caller,
                        project = ds_name,
                        file = ds_callers$file,
                        line = ds_callers$line,
                        source = "r",
                        stringsAsFactors = FALSE))
            }
        }
    }

    # 3. Target-project roxygen @examples
    if ("examples" %in% include && dir.exists(project_dir)) {
        results <- rbind(results, scan_examples(project_dir, fn))
    }

    # 4. Target-project vignettes
    if ("vignettes" %in% include && dir.exists(project_dir)) {
        results <- rbind(results, scan_vignettes(project_dir, fn))
    }

    results
}

#' Empty result frame shared by blast_radius scanners
#' @noRd
empty_blast_results <- function() {
    data.frame(caller = character(), project = character(),
               file = character(), line = integer(), source = character(),
               stringsAsFactors = FALSE)
}

# Tests for blast.R

library(saber)

# --- Setup: create a fake project with functions ---

fake_home <- tempfile("home")
proj <- file.path(fake_home, "mypkg")
dir.create(file.path(proj, "R"), recursive = TRUE)

writeLines(c(
  "helper <- function(x) x + 1",
  "",
  "main_fn <- function(x) {",
  "  helper(x)",
  "}"
), file.path(proj, "R", "code.R"))

writeLines("export(main_fn)", file.path(proj, "NAMESPACE"))

# Create a downstream project that imports mypkg
downstream <- file.path(fake_home, "downstream")
dir.create(file.path(downstream, "R"), recursive = TRUE)

writeLines(c(
  "Package: downstream",
  "Version: 0.1.0",
  "Title: Downstream",
  "Imports: mypkg"
), file.path(downstream, "DESCRIPTION"))

writeLines(c(
  "use_helper <- function() {",
  "  mypkg::helper(42)",
  "}"
), file.path(downstream, "R", "code.R"))

cache <- tempfile("symcache")

# --- blast_radius for an internal function ---

br <- blast_radius("helper", project = proj, scan_dir = fake_home,
                   cache_dir = cache)
expect_true(is.data.frame(br))
expect_true(nrow(br) >= 1L)
expect_true("main_fn" %in% br$caller)

# --- blast_radius finds downstream callers ---

# The downstream project imports mypkg and calls mypkg::helper
expect_true(any(br$project == "downstream"))

# --- blast_radius for a non-existent function ---

br2 <- blast_radius("nonexistent_fn_xyz", project = proj,
                    scan_dir = fake_home, cache_dir = cache)
expect_equal(nrow(br2), 0L)

# --- Cleanup ---
unlink(c(fake_home, cache), recursive = TRUE)

#' @title Code intelligence: function call graph
#' @description Render a project's internal function call graph as interactive
#'   SVG.

#' Render a function call graph for an R project
#'
#' Pulls the AST symbol index via \code{\link{symbols}} and renders
#' internal function-to-function call edges as SVG. External calls
#' (into other packages) are dropped by default.
#'
#' @param project_dir Path to the project directory (an R package root).
#' @param include_external If \code{TRUE}, also include nodes for
#'   functions called from other packages. Default \code{FALSE}.
#' @param ... Passed through to \code{\link{graph_svg}} (e.g.,
#'   \code{width}, \code{height}, \code{iterations}, \code{seed}).
#' @return Character vector of SVG lines. Write with \code{writeLines()}.
#' @examples
#' d <- file.path(tempdir(), "fngdemo")
#' dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
#' writeLines(c("Package: demo", "Version: 0.1.0"),
#'            file.path(d, "DESCRIPTION"))
#' writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
#' writeLines("double <- function(x) add(x, x)",
#'            file.path(d, "R", "double.R"))
#' svg <- fn_graph(d)
#' writeLines(svg, tempfile(fileext = ".svg"))
#' @export
fn_graph <- function(project_dir, include_external = FALSE, ...) {
    idx <- symbols(project_dir)
    defs <- idx$defs
    calls <- idx$calls

    if (!include_external) {
        calls <- calls[calls$callee %in% defs$name,, drop = FALSE]
    }

    node_ids <- unique(c(defs$name, calls$callee, calls$caller))
    node_ids <- node_ids[!is.na(node_ids) & nzchar(node_ids)]

    edges <- data.frame(from = calls$caller, to = calls$callee,
                        stringsAsFactors = FALSE)
    edges <- unique(edges[!is.na(edges$from) & !is.na(edges$to),, drop = FALSE])

    def_match <- match(node_ids, defs$name)
    file_rel <- defs$file[def_match]
    line <- defs$line[def_match]
    exported <- defs$exported[def_match]
    in_deg <- tabulate(match(edges$to, node_ids), nbins = length(node_ids))
    out_deg <- tabulate(match(edges$from, node_ids), nbins = length(node_ids))
    visibility <- ifelse(is.na(exported), "external",
                         ifelse(exported, "exported", "internal"))
    tooltips <- ifelse(
                       is.na(def_match),
                       sprintf("%s\n(external)\ncalled by %d", node_ids, in_deg),
                       sprintf("%s\n%s:%d\n%s\ncalled by %d | calls %d",
                               node_ids, file_rel, line, visibility, in_deg, out_deg))

    nodes <- data.frame(id = node_ids, label = node_ids,
                        href = NA_character_, tooltip = tooltips,
                        stringsAsFactors = FALSE)

    graph_svg(edges, nodes, ...)
}

# Tests for fn_graph.R

library(saber)

d <- file.path(tempdir(), paste0("fngraph-", format(Sys.time(), "%H%M%S")))
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines(c("Package: demo", "Version: 0.1.0"), file.path(d, "DESCRIPTION"))
writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
writeLines("double <- function(x) add(x, x)", file.path(d, "R", "double.R"))

svg <- fn_graph(d, cache_dir = tempdir())
expect_true(is.character(svg))
expect_true(any(grepl("^<svg", svg)))
# Tooltips carry name + file:line + visibility + degree
expect_true(any(grepl("add.R:1", svg, fixed = TRUE)))
expect_true(any(grepl("double.R:1", svg, fixed = TRUE)))
expect_true(any(grepl("called by 1", svg, fixed = TRUE)))
expect_true(any(grepl("calls 1", svg, fixed = TRUE)))
expect_equal(sum(grepl("<line ", svg)), 1L)

# Tests for fn_graph.R

library(saber)

d <- file.path(tempdir(), paste0("fngraph-", format(Sys.time(), "%H%M%S")))
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines(c("Package: demo", "Version: 0.1.0"), file.path(d, "DESCRIPTION"))
writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
writeLines("double <- function(x) add(x, x)", file.path(d, "R", "double.R"))

svg <- fn_graph(d)
expect_true(is.character(svg))
expect_true(any(grepl("^<svg", svg)))
# Tooltips carry name + file:line + visibility + degree
expect_true(any(grepl("add.R:1", svg, fixed = TRUE)))
expect_true(any(grepl("double.R:1", svg, fixed = TRUE)))
expect_true(any(grepl("called by 1", svg, fixed = TRUE)))
expect_true(any(grepl("calls 1", svg, fixed = TRUE)))
expect_equal(sum(grepl("<line ", svg)), 1L)

#' @title Internal utilities
#' @description Shared helper functions for saber.

#' Compute a file hash for change detection
#'
#' @param filepath Path to a file.
#' @return MD5 hash as a hex string.
#' @noRd
file_hash <- function(filepath) {
    tools::md5sum(filepath)[[1L]]
}

#' Default directories to exclude when scanning for projects
#'
#' Returns a character vector of directory basenames that are skipped
#' when scanning for downstream projects. Override by passing a custom
#' \code{exclude} vector to \code{\link{blast_radius}}.
#'
#' @return Character vector of directory basenames.
#' @examples
#' default_exclude()
#' @export
default_exclude <- function() {
    c(
        # User directories
        "Documents", "Downloads", "Desktop", "Music", "Pictures",
        "Videos", "Templates", "Public", "Sync",
        # R internals
        "R", ".Rcheck",
        # Caches and configs
        ".cache", ".local", ".config", ".claude",
        # Build artifacts
        "actions-runner", "node_modules", ".git",
        # Other
        "snap", ".npm", ".cargo", ".rustup"
    )
}

#' @title Agent context assembly
#' @description Assemble context from memory, instructions, and identity files
#'   for AI coding agents.

#' Load context files for an AI coding agent
#'
#' Returns assembled context (memory, project/global instructions, agent
#' identity files) tailored to a specific consumer. Files autoloaded by the
#' agent natively are skipped to avoid duplication.
#'
#' Defaults per agent:
#' \itemize{
#'   \item \code{"claude"} - skips Claude Code project memory and CLAUDE.md
#'     files (autoloaded by 'Claude Code'). Loads Codex memories, AGENTS.md,
#'     USER.md, and SOUL.md when present.
#'   \item \code{"codex"} - skips AGENTS.md and Codex memories (autoloaded by
#'     Codex). Loads Claude Code project memory, CLAUDE.md, USER.md, and
#'     SOUL.md when present.
#'   \item \code{"corteza"} or \code{NULL} - loads everything available.
#' }
#'
#' Project and global instructions are resolved by trying both naming
#' conventions and picking the file relevant to the consumer:
#' \itemize{
#'   \item Project: \code{CLAUDE.md} or \code{AGENTS.md}
#'   \item Global: \code{~/.claude/CLAUDE.md} or
#'     \code{<workspace_dir>/USER.md}
#' }
#'
#' Override the defaults with the \code{include_*} parameters.
#'
#' @param agent Consumer identifier: \code{"claude"}, \code{"codex"},
#'   \code{"corteza"}, or \code{NULL} (interactive / unknown). The legacy
#'   \code{"llamar"} identifier is accepted as an alias for \code{"corteza"}.
#' @param project_dir Project directory to scan for CLAUDE.md / AGENTS.md.
#' @param workspace_dir Optional directory containing SOUL.md and USER.md
#'   (e.g. \code{~/.corteza/workspace}). If \code{NULL}, those files are
#'   skipped.
#' @param memory_base Base directory for 'Claude Code' project memory files.
#' @param claude_global_path Path to the global 'Claude Code' instructions
#'   file. Defaults to \code{~/.claude/CLAUDE.md}.
#' @param include_memory Override default for memory inclusion. Use
#'   \code{TRUE}/\code{FALSE} to force, or \code{NULL} for the agent default.
#' @param include_project Override default for project instructions
#'   (CLAUDE.md / AGENTS.md).
#' @param include_global Override default for global instructions
#'   (~/.claude/CLAUDE.md / USER.md).
#' @param include_soul Override default for SOUL.md inclusion.
#' @param max_memory_lines Maximum lines to include from each memory source.
#' @return Character string of assembled context, or empty string if no
#'   context applies.
#' @examples
#' \donttest{
#' # Codex agent in current project
#' saber::agent_context(agent = "codex")
#'
#' # Corteza with workspace files
#' saber::agent_context(agent = "corteza",
#'                      workspace_dir = "~/.corteza/workspace")
#'
#' # Force-include memory regardless of agent default
#' saber::agent_context(agent = "claude", include_memory = TRUE)
#' }
#' @export
agent_context <- function(agent = NULL, project_dir = getwd(),
                          workspace_dir = NULL,
                          memory_base = file.path(path.expand("~"), ".claude", "projects"),
                          claude_global_path = file.path(path.expand("~"), ".claude", "CLAUDE.md"),
                          include_memory = NULL, include_project = NULL,
                          include_global = NULL, include_soul = NULL,
                          max_memory_lines = 100L) {
    if (is.null(agent)) {
        agent_key <- NA_character_
    } else {
        agent_key <- as.character(agent)[1L]
    }

    defaults <- agent_context_defaults(agent_key)
    incl_mem <- include_memory %||% defaults$memory
    incl_codex_mem <- include_memory %||% defaults$codex_memory
    incl_proj <- include_project %||% defaults$project
    incl_glob <- include_global %||% defaults$global
    incl_soul <- include_soul %||% defaults$soul

    parts <- character(0L)

    if (isTRUE(incl_mem)) {
        mem <- agent_context_memory(project_dir, memory_base, max_memory_lines)
        if (length(mem) > 0L) {
            parts <- c(parts, mem, "")
        }
    }

    if (isTRUE(incl_codex_mem)) {
        codex_mem <- agent_context_codex_memory(max_memory_lines)
        if (length(codex_mem) > 0L) {
            parts <- c(parts, codex_mem, "")
        }
    }

    if (isTRUE(incl_proj)) {
        proj <- agent_context_project(project_dir, agent_key,
                                      forced = !is.null(include_project))
        if (length(proj) > 0L) {
            parts <- c(parts, proj, "")
        }
    }

    if (isTRUE(incl_glob)) {
        glob <- agent_context_global(workspace_dir, agent_key,
                                     claude_global_path,
                                     forced = !is.null(include_global))
        if (length(glob) > 0L) {
            parts <- c(parts, glob, "")
        }
    }

    if (isTRUE(incl_soul) && !is.null(workspace_dir)) {
        soul <- agent_context_soul(workspace_dir)
        if (length(soul) > 0L) {
            parts <- c(parts, soul, "")
        }
    }

    paste(parts, collapse = "\n")
}

#' Per-agent default inclusion flags
#' @noRd
agent_context_defaults <- function(agent) {
    if (is.na(agent)) {
        return(list(memory = TRUE, codex_memory = TRUE, project = TRUE,
                    global = TRUE, soul = TRUE))
    }
    switch(agent,
           claude = list(memory = FALSE, codex_memory = TRUE, project = TRUE,
                         global = TRUE, soul = TRUE),
           codex = list(memory = TRUE, codex_memory = FALSE, project = TRUE,
                        global = TRUE, soul = TRUE),
           corteza = list(memory = TRUE, codex_memory = TRUE, project = TRUE,
                          global = TRUE, soul = TRUE),
           llamar = list(memory = TRUE, codex_memory = TRUE, project = TRUE,
                         global = TRUE, soul = TRUE),
           list(memory = TRUE, codex_memory = TRUE, project = TRUE,
                global = TRUE, soul = TRUE)
    )
}

#' Load project memory section ('Claude Code' per-project memory)
#' @noRd
agent_context_memory <- function(project_dir, memory_base, max_lines) {
    if (is.null(memory_base) || !dir.exists(memory_base)) {
        return(character(0L))
    }

    project <- basename(normalizePath(project_dir, mustWork = FALSE))
    mem_file <- NULL
    mem_dirs <- list.dirs(memory_base, recursive = FALSE, full.names = TRUE)
    for (md in mem_dirs) {
        proj_encoded <- basename(md)
        proj_name <- sub("^.*-home-[^-]+-", "", proj_encoded)
        if (proj_name == project) {
            candidate <- file.path(md, "memory", "MEMORY.md")
            if (file.exists(candidate)) {
                mem_file <- candidate
                break
            }
        }
    }

    if (is.null(mem_file)) {
        return(character(0L))
    }

    mem_lines <- readLines(mem_file, warn = FALSE)
    lines <- "## Memory"
    if (length(mem_lines) > max_lines) {
        lines <- c(lines, mem_lines[seq_len(max_lines)],
                   sprintf("_... truncated (%d more lines)_",
                           length(mem_lines) - max_lines))
    } else {
        lines <- c(lines, mem_lines)
    }
    lines
}

#' Load Codex memories for non-Codex agents
#' @noRd
agent_context_codex_memory <- function(max_lines) {
    mem_dir <- agent_context_codex_memory_dir()
    if (!dir.exists(mem_dir)) {
        return(character(0L))
    }

    files <- list.files(mem_dir, recursive = TRUE, full.names = TRUE)
    if (length(files) == 0L) {
        return(character(0L))
    }

    info <- file.info(files)
    keep <- !is.na(info$isdir) & !info$isdir & !is.na(info$size) &
    info$size > 0L
    files <- sort(files[keep])
    if (length(files) == 0L) {
        return(character(0L))
    }

    max_lines <- max(1L, as.integer(max_lines)[1L])
    remaining <- max_lines
    out <- "## Codex Memories"
    truncated <- FALSE

    for (i in seq_along(files)) {
        if (remaining <= 0L) {
            truncated <- TRUE
            break
        }

        content <- tryCatch(readLines(files[[i]], warn = FALSE),
                            error = function(e) character(0L))
        if (length(content) == 0L) {
            next
        }

        take <- min(length(content), remaining)
        label <- substring(files[[i]], nchar(mem_dir) + 2L)
        label <- gsub("\\\\", "/", label)
        out <- c(out, "", sprintf("### %s", label), "", content[seq_len(take)])
        remaining <- remaining - take

        if (take < length(content)) {
            truncated <- TRUE
            break
        }
    }

    if (length(out) == 1L) {
        return(character(0L))
    }
    if (truncated) {
        out <- c(out, sprintf("_... truncated after %d lines_", max_lines))
    }
    out
}

#' Resolve the Codex memory directory
#' @noRd
agent_context_codex_memory_dir <- function() {
    codex_home <- Sys.getenv("CODEX_HOME", unset = "")
    if (nchar(codex_home) == 0L) {
        codex_home <- file.path(path.expand("~"), ".codex")
    }
    file.path(path.expand(codex_home), "memories")
}

#' Resolve and load project instructions (CLAUDE.md or AGENTS.md)
#'
#' Picks the file the consumer doesn't already autoload. Ties broken by
#' preferring CLAUDE.md.
#' @noRd
agent_context_project <- function(project_dir, agent, forced = FALSE) {
    claude_path <- file.path(project_dir, "CLAUDE.md")
    agents_path <- file.path(project_dir, "AGENTS.md")
    claude_exists <- file.exists(claude_path)
    agents_exists <- file.exists(agents_path)

    if (!claude_exists && !agents_exists) {
        return(character(0L))
    }

    file_to_load <- NULL
    if (forced || is.na(agent)) {
        # User overrode the default, or unknown agent: prefer CLAUDE.md
        if (claude_exists) {
            file_to_load <- claude_path
        } else {
            file_to_load <- agents_path
        }
    } else if (identical(agent, "claude")) {
        # claude autoloads CLAUDE.md; only load AGENTS.md if it exists
        # and is a distinct file
        if (agents_exists && !same_file(claude_path, agents_path)) {
            file_to_load <- agents_path
        }
    } else if (identical(agent, "codex")) {
        # codex autoloads AGENTS.md; only load CLAUDE.md if it exists
        # and is a distinct file
        if (claude_exists && !same_file(claude_path, agents_path)) {
            file_to_load <- claude_path
        }
    } else {
        # corteza / legacy aliases / unknown: prefer CLAUDE.md, fall back to AGENTS.md
        if (claude_exists) {
            file_to_load <- claude_path
        } else {
            file_to_load <- agents_path
        }
    }

    if (is.null(file_to_load)) {
        return(character(0L))
    }

    content <- tryCatch(readLines(file_to_load, warn = FALSE),
                        error = function(e) character(0L))
    if (length(content) == 0L) {
        return(character(0L))
    }

    c(sprintf("## %s", basename(file_to_load)), "", content)
}

#' Resolve and load global instructions
#'
#' Picks claude_global_path or workspace_dir/USER.md based on consumer.
#' @noRd
agent_context_global <- function(workspace_dir, agent, claude_global,
                                 forced = FALSE) {
    user_path <- if (!is.null(workspace_dir)) {
        file.path(workspace_dir, "USER.md")
    } else {
        NULL
    }

    claude_exists <- file.exists(claude_global)
    user_exists <- !is.null(user_path) && file.exists(user_path)

    if (!claude_exists && !user_exists) {
        return(character(0L))
    }

    file_to_load <- NULL
    if (forced || is.na(agent)) {
        if (claude_exists) {
            file_to_load <- claude_global
        } else {
            file_to_load <- user_path
        }
    } else if (identical(agent, "claude")) {
        # claude autoloads ~/.claude/CLAUDE.md; only load USER.md
        if (user_exists && !same_file(claude_global, user_path)) {
            file_to_load <- user_path
        }
    } else {
        # codex / corteza / legacy aliases / unknown: prefer claude global,
        # fall back to USER.md
        if (claude_exists) {
            file_to_load <- claude_global
        } else {
            file_to_load <- user_path
        }
    }

    if (is.null(file_to_load)) {
        return(character(0L))
    }

    content <- tryCatch(readLines(file_to_load, warn = FALSE),
                        error = function(e) character(0L))
    if (length(content) == 0L) {
        return(character(0L))
    }

    label <- if (identical(file_to_load, claude_global)) {
        "Global Instructions (~/.claude/CLAUDE.md)"
    } else {
        "User Preferences (USER.md)"
    }
    c(sprintf("## %s", label), "", content)
}

#' Load SOUL.md from workspace
#' @noRd
agent_context_soul <- function(workspace_dir) {
    soul_path <- file.path(workspace_dir, "SOUL.md")
    if (!file.exists(soul_path)) {
        return(character(0L))
    }
    content <- tryCatch(readLines(soul_path, warn = FALSE),
                        error = function(e) character(0L))
    if (length(content) == 0L) {
        return(character(0L))
    }
    c("## Agent Identity (SOUL.md)", "", content)
}

#' Check if two paths resolve to the same file
#'
#' Returns FALSE if either path doesn't exist.
#' @noRd
same_file <- function(a, b) {
    if (is.null(a) || is.null(b)) {
        return(FALSE)
    }
    if (!file.exists(a) || !file.exists(b)) {
        return(FALSE)
    }
    norm_a <- tryCatch(normalizePath(a, mustWork = FALSE),
                       error = function(e) a)
    norm_b <- tryCatch(normalizePath(b, mustWork = FALSE),
                       error = function(e) b)
    identical(norm_a, norm_b)
}

# Tests for agent_context()

# --- Setup ---
root <- file.path(tempdir(), "test_agent_context")
unlink(root, recursive = TRUE)
dir.create(root, recursive = TRUE, showWarnings = FALSE)

project_dir <- file.path(root, "demopkg")
dir.create(project_dir, recursive = TRUE, showWarnings = FALSE)

workspace_dir <- file.path(root, "workspace")
dir.create(workspace_dir, recursive = TRUE, showWarnings = FALSE)

mem_base <- file.path(root, "claude_mem")
mem_proj_dir <- file.path(mem_base, "-home-user-demopkg", "memory")
dir.create(mem_proj_dir, recursive = TRUE, showWarnings = FALSE)

codex_home <- file.path(root, "codex_home")
codex_mem_dir <- file.path(codex_home, "memories")
dir.create(codex_mem_dir, recursive = TRUE, showWarnings = FALSE)

# Fake claude global path (points to nonexistent file by default).
# Use this in all tests to avoid leaking the real ~/.claude/CLAUDE.md.
fake_claude_global <- file.path(root, "fake_home_claude.md")

# Helper that injects the test fixtures
ac <- function(agent, workspace = workspace_dir, ...) {
    old_codex_home <- Sys.getenv("CODEX_HOME", unset = NA_character_)
    Sys.setenv(CODEX_HOME = codex_home)
    on.exit({
        if (is.na(old_codex_home)) {
            Sys.unsetenv("CODEX_HOME")
        } else {
            Sys.setenv(CODEX_HOME = old_codex_home)
        }
    }, add = TRUE)

    saber::agent_context(agent = agent,
                         project_dir = project_dir,
                         workspace_dir = workspace,
                         memory_base = mem_base,
                         claude_global_path = fake_claude_global,
                         ...)
}

write_lines_to <- function(path, lines) {
    writeLines(lines, path)
}

# --- Empty case: no files at all ---
expect_equal(ac("corteza"), "")

# --- AGENTS.md only, codex agent (codex autoloads it -> skipped) ---
write_lines_to(file.path(project_dir, "AGENTS.md"), "Project rules.")
result <- ac("codex")
expect_false(grepl("AGENTS.md", result))
expect_false(grepl("Project rules", result))

# --- AGENTS.md only, claude agent (claude doesn't autoload AGENTS.md) ---
result <- ac("claude")
expect_true(grepl("AGENTS.md", result))
expect_true(grepl("Project rules", result))

# --- CLAUDE.md only, claude agent (autoloaded -> skipped) ---
file.remove(file.path(project_dir, "AGENTS.md"))
write_lines_to(file.path(project_dir, "CLAUDE.md"), "Claude project rules.")
result <- ac("claude")
expect_false(grepl("Claude project rules", result))

# --- CLAUDE.md only, codex agent (loads it as fallback) ---
result <- ac("codex")
expect_true(grepl("CLAUDE.md", result))
expect_true(grepl("Claude project rules", result))

# --- Both files exist, corteza prefers CLAUDE.md ---
write_lines_to(file.path(project_dir, "AGENTS.md"), "Agents version.")
result <- ac("corteza")
expect_true(grepl("Claude project rules", result))
expect_false(grepl("Agents version", result))

# --- Memory loading ---
write_lines_to(file.path(mem_proj_dir, "MEMORY.md"),
               c("- [Test](t.md) - a test memory"))
write_lines_to(file.path(codex_mem_dir, "reciprocal.md"),
               "saber is meant to be reciprocal")
result <- ac("corteza")
expect_true(grepl("## Memory", result))
expect_true(grepl("test memory", result))
expect_true(grepl("## Codex Memories", result))
expect_true(grepl("saber is meant to be reciprocal", result))

# --- Claude memory skipped for claude by default, Codex memory is included ---
result <- ac("claude")
expect_false(grepl("## Memory", result))
expect_true(grepl("saber is meant to be reciprocal", result))

# --- Codex memory skipped for codex by default, Claude memory is included ---
result <- ac("codex")
expect_true(grepl("test memory", result))
expect_false(grepl("saber is meant to be reciprocal", result))

# --- Corteza receives both memory systems ---
result <- ac("corteza")
expect_true(grepl("test memory", result))
expect_true(grepl("saber is meant to be reciprocal", result))

# --- Memory force-included for claude includes both memory systems ---
result <- ac("claude", include_memory = TRUE)
expect_true(grepl("## Memory", result))
expect_true(grepl("saber is meant to be reciprocal", result))

# --- Memory can be disabled for all memory systems ---
result <- ac("corteza", include_memory = FALSE)
expect_false(grepl("test memory", result))
expect_false(grepl("saber is meant to be reciprocal", result))

# --- Legacy llamar identifier remains a corteza alias ---
result <- ac("llamar")
expect_true(grepl("test memory", result))
expect_true(grepl("saber is meant to be reciprocal", result))

# --- SOUL.md and USER.md from workspace ---
write_lines_to(file.path(workspace_dir, "SOUL.md"), "I am the soul.")
write_lines_to(file.path(workspace_dir, "USER.md"), "User preferences here.")
result <- ac("corteza")
expect_true(grepl("SOUL.md", result))
expect_true(grepl("I am the soul", result))

# --- USER.md loads for claude when claude_global_path doesn't exist ---
result <- ac("claude")
expect_true(grepl("User preferences here", result))

# --- claude_global_path file is loaded for codex/corteza ---
write_lines_to(fake_claude_global, "Fake claude global content.")
result <- ac("corteza")
expect_true(grepl("Fake claude global content", result))
# When claude global exists, USER.md is the fallback (not loaded)
expect_false(grepl("User preferences here", result))

# --- claude_global_path is skipped for claude (autoloaded) ---
result <- ac("claude")
expect_false(grepl("Fake claude global content", result))
# But USER.md still loads for claude
expect_true(grepl("User preferences here", result))

# Reset for remaining tests
file.remove(fake_claude_global)

# --- include_soul = FALSE skips SOUL.md ---
result <- ac("corteza", include_soul = FALSE)
expect_false(grepl("I am the soul", result))

# --- workspace_dir = NULL skips SOUL.md and USER.md ---
result <- ac("corteza", workspace = NULL)
expect_false(grepl("SOUL", result))
expect_false(grepl("USER", result))

# --- NULL agent loads everything ---
result <- ac(NULL)
expect_true(grepl("## Memory", result))
expect_true(grepl("Claude project rules", result))
expect_true(grepl("I am the soul", result))

# --- include_project = FALSE skips even when files exist ---
result <- ac("corteza", include_project = FALSE)
expect_false(grepl("Claude project rules", result))

# --- Symlink dedup: AGENTS.md symlink to CLAUDE.md, claude agent skips both ---
file.remove(file.path(project_dir, "AGENTS.md"))
file.symlink("CLAUDE.md", file.path(project_dir, "AGENTS.md"))
result <- ac("claude")
# AGENTS.md is a symlink to CLAUDE.md, and claude autoloads CLAUDE.md,
# so AGENTS.md should also be skipped (same canonical file)
expect_false(grepl("Claude project rules", result))

# --- Cleanup ---
unlink(root, recursive = TRUE)

#' @title Project discovery
#' @description Discover R package projects and map their dependency
#'   relationships.

#' Discover R package projects
#'
#' Scans a directory for subdirectories containing a DESCRIPTION file and
#' returns their metadata.
#'
#' @param scan_dir Directory to scan for project directories.
#' @param exclude Character vector of directory basenames to skip.
#' @return A data.frame with columns: package, title, version, path, depends,
#'   imports.
#' @examples
#' d <- file.path(tempdir(), "scandir")
#' dir.create(d, showWarnings = FALSE)
#' pkg <- file.path(d, "mypkg")
#' dir.create(pkg, showWarnings = FALSE)
#' writeLines(c("Package: mypkg", "Title: Demo", "Version: 0.1.0"),
#'            file.path(pkg, "DESCRIPTION"))
#' projects(scan_dir = d)
#' @export
projects <- function(scan_dir = path.expand("~"), exclude = default_exclude()) {
    project_dirs <- list.dirs(scan_dir, recursive = FALSE, full.names = TRUE)
    project_dirs <- project_dirs[!basename(project_dirs) %in% exclude]

    rows <- list()
    for (d in project_dirs) {
        desc_file <- file.path(d, "DESCRIPTION")
        if (!file.exists(desc_file)) {
            next
        }

        dcf <- tryCatch(
                        read.dcf(desc_file,
                                 fields = c("Package", "Title", "Version", "Depends",
                    "Imports", "LinkingTo")),
                        error = function(e) NULL
        )
        if (is.null(dcf) || nrow(dcf) == 0L) {
            next
        }

        pkg_name <- dcf[1L, "Package"]
        if (is.na(pkg_name) || nchar(trimws(pkg_name)) == 0L) {
            next
        }

        rows[[length(rows) + 1L]] <- data.frame(
            package = pkg_name,
            title = na_to_empty(dcf[1L, "Title"]),
            version = na_to_empty(dcf[1L, "Version"]),
            path = d,
            depends = na_to_empty(dcf[1L, "Depends"]),
            imports = na_to_empty(dcf[1L, "Imports"]),
            stringsAsFactors = FALSE
        )
    }

    if (length(rows) == 0L) {
        return(data.frame(package = character(), title = character(),
                          version = character(), path = character(),
                          depends = character(), imports = character(),
                          stringsAsFactors = FALSE))
    }
    do.call(rbind, rows)
}

#' Find projects that depend on a given package
#'
#' Scans DESCRIPTION files in project directories under \code{scan_dir}
#' for Depends, Imports, or LinkingTo fields that reference \code{package}.
#'
#' @param package Character. Package name to search for.
#' @param scan_dir Directory to scan for project directories.
#' @param exclude Character vector of directory basenames to skip.
#' @return Character vector of project names that depend on \code{package}.
#' @examples
#' d <- file.path(tempdir(), "dsdir")
#' dir.create(d, showWarnings = FALSE)
#' pkg <- file.path(d, "child")
#' dir.create(pkg, showWarnings = FALSE)
#' writeLines(c("Package: child", "Version: 0.1.0", "Imports: parent"),
#'            file.path(pkg, "DESCRIPTION"))
#' find_downstream("parent", scan_dir = d)
#' @export
find_downstream <- function(package, scan_dir = path.expand("~"),
                            exclude = default_exclude()) {
    project_dirs <- list.dirs(scan_dir, recursive = FALSE, full.names = TRUE)
    project_dirs <- project_dirs[!basename(project_dirs) %in% exclude]
    downstream <- character(0L)

    for (d in project_dirs) {
        desc_file <- file.path(d, "DESCRIPTION")
        if (!file.exists(desc_file)) {
            next
        }

        dcf <- tryCatch(
                        read.dcf(desc_file, fields = c("Depends", "Imports", "LinkingTo")),
                        error = function(e) NULL
        )
        if (is.null(dcf) || nrow(dcf) == 0L) {
            next
        }

        deps <- character(0L)
        for (field in c("Depends", "Imports", "LinkingTo")) {
            deps <- c(deps, parse_dcf_list(dcf[1L, field]))
        }
        if (package %in% deps) {
            downstream <- c(downstream, basename(d))
        }
    }

    downstream
}

#' Replace NA with empty string
#' @noRd
na_to_empty <- function(x) {
    if (is.na(x)) {
        ""
    } else {
        trimws(x)
    }
}

#' Parse a comma-separated DCF field into a clean character vector
#' @noRd
parse_dcf_list <- function(x) {
    if (is.na(x) || nchar(trimws(x)) == 0L) {
        return(character(0L))
    }
    parts <- strsplit(x, ",")[[1L]]
    parts <- trimws(parts)
    parts <- sub("\\s*\\(.*\\)", "", parts)
    parts <- parts[nchar(parts) > 0L]
    parts
}

# Tests for projects() and find_downstream()

# --- Setup ---
scan_dir <- file.path(tempdir(), "test_projects")
dir.create(scan_dir, showWarnings = FALSE)

pkg_a <- file.path(scan_dir, "pkgA")
pkg_b <- file.path(scan_dir, "pkgB")
dir.create(pkg_a, showWarnings = FALSE)
dir.create(pkg_b, showWarnings = FALSE)

writeLines(c(
    "Package: pkgA",
    "Title: Package A",
    "Version: 1.0.0",
    "Imports: stats"
), file.path(pkg_a, "DESCRIPTION"))

writeLines(c(
    "Package: pkgB",
    "Title: Package B",
    "Version: 0.2.0",
    "Imports: pkgA, utils"
), file.path(pkg_b, "DESCRIPTION"))

# --- projects() ---
res <- projects(scan_dir = scan_dir)
expect_true(is.data.frame(res))
expect_equal(nrow(res), 2L)
expect_true("pkgA" %in% res$package)
expect_true("pkgB" %in% res$package)
expect_true("package" %in% names(res))
expect_true("title" %in% names(res))
expect_true("version" %in% names(res))
expect_true("path" %in% names(res))

row_a <- res[res$package == "pkgA", ]
expect_equal(row_a$title, "Package A")
expect_equal(row_a$version, "1.0.0")

# --- projects() empty dir ---
empty <- file.path(tempdir(), "test_projects_empty")
dir.create(empty, showWarnings = FALSE)
res_empty <- projects(scan_dir = empty)
expect_equal(nrow(res_empty), 0L)

# --- find_downstream() ---
ds <- find_downstream("pkgA", scan_dir = scan_dir)
expect_equal(ds, "pkgB")

ds_none <- find_downstream("pkgB", scan_dir = scan_dir)
expect_equal(length(ds_none), 0L)

ds_missing <- find_downstream("nonexistent", scan_dir = scan_dir)
expect_equal(length(ds_missing), 0L)

# --- Cleanup ---
unlink(scan_dir, recursive = TRUE)
unlink(empty, recursive = TRUE)

#!/usr/bin/env Rscript
# saber - generate project context at session start
# For use as a Claude Code or Codex SessionStart hook
#
# Usage: Rscript session-start.R [agent]
#   agent: "claude", "codex", or omit for interactive default

cli_args <- commandArgs(trailingOnly = TRUE)
agent <- if (length(cli_args) > 0L) cli_args[[1L]] else NULL

session_cwd <- getwd()

global_preferences_path <- function() {
    custom <- Sys.getenv("AGENTS_GLOBAL_MD", unset = "")
    if (nchar(custom) > 0L) {
        return(path.expand(custom))
    }
    path.expand("~/.config/agents/GLOBAL.md")
}

load_global_preferences <- function() {
    path <- global_preferences_path()
    if (!file.exists(path)) {
        return(NULL)
    }

    lines <- tryCatch(readLines(path, warn = FALSE),
                      error = function(e) NULL)
    if (is.null(lines)) {
        return(NULL)
    }

    text <- paste(lines, collapse = "\n")
    if (nchar(trimws(text)) == 0L) {
        return(NULL)
    }

    paste0("## Global Preferences\n\n", text, "\n")
}

resolve_repo_root <- function(path) {
    root <- tryCatch(
                    system2("git", c("-C", path, "rev-parse", "--show-toplevel"),
                            stdout = TRUE, stderr = FALSE),
                    error = function(e) character(0L)
    )
    if (length(root) == 0L) {
        return(NULL)
    }

    root <- trimws(root[[1L]])
    if (nchar(root) == 0L) {
        return(NULL)
    }
    path.expand(root)
}

load_saber_fun <- function(name, repo_root = NULL) {
    if (!is.null(repo_root)) {
        local_fun <- tryCatch(
                            {
                                r_dir <- file.path(repo_root, "R")
                                if (!file.exists(file.path(repo_root, "DESCRIPTION")) ||
                                    !dir.exists(r_dir)) {
                                    stop("No local package source found")
                                }

                                env <- new.env(parent = baseenv())
                                r_files <- list.files(r_dir, pattern = "\\.[Rr]$",
                                                      full.names = TRUE)
                                for (f in r_files) {
                                    sys.source(f, envir = env)
                                }
                                get(name, envir = env, inherits = FALSE)
                            },
                            error = function(e) NULL
        )
        if (is.function(local_fun)) {
            return(local_fun)
        }
    }

    if (requireNamespace("saber", quietly = TRUE)) {
        return(getExportedValue("saber", name))
    }

    stop("saber not available")
}

load_agent_memory <- function(agent, project_dir, repo_root = NULL) {
    context_fun <- load_saber_fun("agent_context", repo_root)
    context_fun(agent = agent, project_dir = project_dir,
                include_project = FALSE,
                include_global = FALSE,
                include_soul = FALSE)
}

append_context <- function(text, section) {
    if (is.null(section) || nchar(trimws(section)) == 0L) {
        return(text)
    }
    paste0(text, "\n\n", section)
}

repo_root <- resolve_repo_root(session_cwd)
if (!is.null(repo_root)) {
    project <- basename(repo_root)
    scan_dir <- dirname(repo_root)
    project_dir <- repo_root
} else {
    project <- basename(session_cwd)
    scan_dir <- path.expand("~")
    project_dir <- session_cwd
}

briefing_text <- tryCatch(
    {
        briefing_fun <- load_saber_fun("briefing", repo_root)
        msg_lines <- utils::capture.output(
            briefing_fun(project, scan_dir = scan_dir),
            type = "message"
        )
        paste(msg_lines, collapse = "\n")
    },
    error = function(e) {
        paste0("# Briefing: ", project,
               "\n_saber not available:_ ", conditionMessage(e), "\n")
    }
)

if (is.null(briefing_text) || nchar(briefing_text) == 0L) {
    briefing_text <- paste0("# Briefing: ", project,
                            "\n_No briefing available._\n")
}

memory_text <- tryCatch(
    load_agent_memory(agent, project_dir, repo_root),
    error = function(e) NULL
)
briefing_text <- append_context(briefing_text, memory_text)

global_preferences <- load_global_preferences()
if (!is.null(global_preferences)) {
    briefing_text <- append_context(briefing_text, global_preferences)
}

escaped <- gsub("\\\\", "\\\\\\\\", briefing_text)
escaped <- gsub("\"", "\\\\\"", escaped)
escaped <- gsub("\n", "\\\\n", escaped)
escaped <- gsub("\t", "\\\\t", escaped)

cat(sprintf('{
  "hookSpecificOutput": {
    "hookEventName": "SessionStart",
    "additionalContext": "%s"
  }
}', escaped))

#' @title Project discovery: package dependency graph
#' @description Render the dependency graph across a set of R packages as
#'   interactive SVG.

#' Render a package-level dependency graph
#'
#' Discovers R packages under \code{scan_dir} via \code{\link{projects}},
#' parses each one's \code{Depends} and \code{Imports} fields, and
#' renders edges between packages that both live in \code{scan_dir}.
#' External CRAN dependencies are dropped.
#'
#' @param scan_dir Directory to scan for project directories.
#' @param packages Optional character vector limiting the graph to
#'   these packages.
#' @param include_suggests If \code{TRUE}, also include edges for
#'   packages in each project's \code{Suggests} field. Default
#'   \code{FALSE} (only \code{Depends} and \code{Imports}).
#' @param ... Passed through to \code{\link{graph_svg}}.
#' @return Character vector of SVG lines. Write with \code{writeLines()}.
#' @examples
#' d <- file.path(tempdir(), "pkgdemo")
#' dir.create(file.path(d, "parent"), recursive = TRUE, showWarnings = FALSE)
#' dir.create(file.path(d, "child"), showWarnings = FALSE)
#' writeLines(c("Package: parent", "Title: P", "Version: 0.1.0"),
#'            file.path(d, "parent", "DESCRIPTION"))
#' writeLines(c("Package: child", "Title: C", "Version: 0.1.0",
#'              "Imports: parent"),
#'            file.path(d, "child", "DESCRIPTION"))
#' svg <- pkg_graph(scan_dir = d)
#' writeLines(svg, tempfile(fileext = ".svg"))
#' @export
pkg_graph <- function(scan_dir = path.expand("~"), packages = NULL,
                      include_suggests = FALSE, ...) {
    projs <- projects(scan_dir = scan_dir)
    if (!is.null(packages)) {
        projs <- projs[projs$package %in% packages,, drop = FALSE]
    }
    if (nrow(projs) == 0L) {
        stop("No packages found under ", scan_dir)
    }

    known <- projs$package
    edges <- list()
    dep_counts <- integer(nrow(projs))
    local_counts <- integer(nrow(projs))
    for (i in seq_len(nrow(projs))) {
        all_deps <- c(parse_deps(projs$depends[i]),
                      parse_deps(projs$imports[i]))
        if (include_suggests) {
            all_deps <- c(all_deps, read_suggests(projs$path[i]))
        }
        local <- intersect(all_deps, known)
        dep_counts[i] <- length(unique(all_deps))
        local_counts[i] <- length(local)
        if (length(local)) {
            edges[[i]] <- data.frame(from = projs$package[i], to = local,
                                     stringsAsFactors = FALSE)
        }
    }
    edges <- do.call(rbind, edges)
    if (is.null(edges)) {
        edges <- data.frame(from = character(), to = character(),
                            stringsAsFactors = FALSE)
    }

    tooltips <- sprintf("%s %s\n%s\n%d deps (%d local)",
                        projs$package, projs$version,
                        ifelse(is.na(projs$title) | !nzchar(projs$title),
                               "(no title)", projs$title),
                        dep_counts, local_counts)

    nodes <- data.frame(id = projs$package, label = projs$package,
                        href = NA_character_, tooltip = tooltips,
                        stringsAsFactors = FALSE)

    graph_svg(edges, nodes, ...)
}

#' Read the Suggests field from a package's DESCRIPTION, returning a
#' character vector of package names with version constraints stripped.
#' @noRd
read_suggests <- function(pkg_path) {
    desc_path <- file.path(pkg_path, "DESCRIPTION")
    if (!file.exists(desc_path)) {
        return(character())
    }
    fields <- read.dcf(desc_path, fields = "Suggests")
    parse_deps(fields[1L, "Suggests"])
}

#' Parse a comma-separated dependency field from a DESCRIPTION column,
#' stripping version constraints and whitespace.
#' @noRd
parse_deps <- function(s) {
    if (is.null(s) || is.na(s) || !nzchar(s)) {
        return(character())
    }
    parts <- strsplit(s, ",", fixed = TRUE)[[1L]]
    parts <- sub("\\s*\\(.*\\)\\s*$", "", parts)
    parts <- trimws(parts)
    parts <- parts[parts != "R" & nzchar(parts)]
    parts
}

# Tests for pkg_graph.R

library(saber)

d <- file.path(tempdir(), paste0("pkggraph-", format(Sys.time(), "%H%M%S")))
dir.create(d, showWarnings = FALSE)

parent_dir <- file.path(d, "parent")
child_dir <- file.path(d, "child")
dir.create(parent_dir, showWarnings = FALSE)
dir.create(child_dir, showWarnings = FALSE)
writeLines(c("Package: parent", "Title: Parent", "Version: 0.1.0"),
           file.path(parent_dir, "DESCRIPTION"))
writeLines(c("Package: child", "Title: Child", "Version: 0.1.0",
             "Imports: parent (>= 0.1.0)"),
           file.path(child_dir, "DESCRIPTION"))

svg <- pkg_graph(scan_dir = d)
# Tooltips carry title + version + dep counts
expect_true(any(grepl("parent 0.1.0", svg, fixed = TRUE)))
expect_true(any(grepl("child 0.1.0", svg, fixed = TRUE)))
expect_true(any(grepl("1 deps (1 local)", svg, fixed = TRUE)))
expect_equal(sum(grepl("<line ", svg)), 1L)

expect_equal(saber:::parse_deps("foo (>= 1.0), bar, R (>= 3.5)"),
             c("foo", "bar"))
expect_equal(saber:::parse_deps(""), character())
expect_equal(saber:::parse_deps(NA_character_), character())

svg <- pkg_graph(scan_dir = d, packages = "parent")
expect_true(any(grepl("parent 0.1.0", svg, fixed = TRUE)))
expect_false(any(grepl("child 0.1.0", svg, fixed = TRUE)))

#' @title Force-directed graph rendering
#' @description Render a graph as static, interactive SVG using a base R
#'   Fruchterman-Reingold force simulation.

#' Render a graph as static SVG
#'
#' Runs a Fruchterman-Reingold force simulation to lay out nodes, then
#' emits SVG with baked coordinates. Output is interactive via native
#' browser features: hover tooltips from \code{<title>} elements, click
#' navigation from \code{<a xlink:href>} wrappers, CSS \code{:hover}
#' highlighting. No JavaScript.
#'
#' Suitable for graphs up to roughly a few hundred nodes. The repulsion
#' step is vectorized via \code{outer()} but allocates an
#' \code{n x n} matrix per iteration; larger graphs should pre-filter.
#'
#' @param edges Data frame with \code{from} and \code{to} columns
#'   holding node ids.
#' @param nodes Optional data frame with \code{id}, \code{label},
#'   \code{href} columns. Optionally a \code{tooltip} column (plain
#'   text, may contain newlines) that overrides the default hover
#'   text (which is the label). \code{id} must cover every node
#'   mentioned in \code{edges}. If \code{NULL}, ids from \code{edges}
#'   are used as labels with no hrefs or tooltips.
#' @param width SVG viewport width in pixels.
#' @param height SVG viewport height in pixels.
#' @param iterations Force-simulation steps. 50 is usually enough.
#' @param seed Integer seed for the random initial layout (output is
#'   deterministic given the same seed).
#' @return Character vector, one SVG element per line. Write with
#'   \code{writeLines()}.
#' @importFrom stats runif
#' @importFrom utils head
#' @examples
#' edges <- data.frame(from = c("a", "a", "b"),
#'                     to = c("b", "c", "c"))
#' svg <- graph_svg(edges)
#' writeLines(svg, tempfile(fileext = ".svg"))
#' @export
graph_svg <- function(edges, nodes = NULL, width = 1200L, height = 900L,
                      iterations = 50L, seed = 1L) {
    ids <- unique(c(edges$from, edges$to))
    if (is.null(nodes)) {
        nodes <- data.frame(id = ids, label = ids, href = NA_character_,
                            stringsAsFactors = FALSE)
    }
    if (!all(ids %in% nodes$id)) {
        missing <- setdiff(ids, nodes$id)
        stop("Edge refers to unknown node id: ",
             paste(head(missing, 5L), collapse = ", "))
    }

    n <- nrow(nodes)
    from_i <- match(edges$from, nodes$id)
    to_i <- match(edges$to, nodes$id)

    set.seed(seed)
    x <- runif(n, width / 4, 3 * width / 4)
    y <- runif(n, height / 4, 3 * height / 4)
    # FR's ideal edge length. Cap at min-dim/4 so small graphs don't
    # stretch to the viewport bounds.
    k <- min(sqrt(width * height / max(n, 1L)), min(width, height) / 4)
    temp <- width / 20
    cx <- width / 2
    cy <- height / 2
    # Gravity pulls nodes toward the center each iteration. Without it,
    # repulsion wins and nodes pile up on the clamp at the edges.
    gravity <- 0.05
    margin <- 40

    for (iter in seq_len(iterations)) {
        dx_m <- outer(x, x, "-")
        dy_m <- outer(y, y, "-")
        dist_m <- sqrt(dx_m ^ 2 + dy_m ^ 2) + 1e-6
        rep_f <- k ^ 2 / dist_m
        diag(rep_f) <- 0
        dx <- rowSums(dx_m / dist_m * rep_f)
        dy <- rowSums(dy_m / dist_m * rep_f)

        if (length(from_i)) {
            evx <- x[from_i] - x[to_i]
            evy <- y[from_i] - y[to_i]
            edist <- sqrt(evx ^ 2 + evy ^ 2) + 1e-6
            att_f <- edist ^ 2 / k
            ax <- evx / edist * att_f
            ay <- evy / edist * att_f
            for (e in seq_along(from_i)) {
                dx[from_i[e]] <- dx[from_i[e]] - ax[e]
                dy[from_i[e]] <- dy[from_i[e]] - ay[e]
                dx[to_i[e]] <- dx[to_i[e]] + ax[e]
                dy[to_i[e]] <- dy[to_i[e]] + ay[e]
            }
        }

        dx <- dx - (x - cx) * gravity * k
        dy <- dy - (y - cy) * gravity * k

        disp <- sqrt(dx ^ 2 + dy ^ 2) + 1e-6
        x <- x + dx / disp * pmin(disp, temp)
        y <- y + dy / disp * pmin(disp, temp)
        x <- pmin(pmax(x, margin), width - margin)
        y <- pmin(pmax(y, margin), height - margin)
        temp <- temp * 0.95
    }

    out <- c(
             sprintf('<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink" viewBox="0 0 %d %d">',
                     width, height),
             "<style>",
             "  .node { fill: #4c78a8; stroke: #fff; stroke-width: 1.5; cursor: pointer; }",
             "  .node:hover { fill: #f58518; }",
             "  .edge { stroke: #999; stroke-opacity: 0.4; }",
             "  .label { font: 11px sans-serif; pointer-events: none; fill: #333; }",
             "</style>"
    )

    for (e in seq_along(from_i)) {
        out <- c(out, sprintf(
                              '<line class="edge" x1="%.1f" y1="%.1f" x2="%.1f" y2="%.1f"/>',
                              x[from_i[e]], y[from_i[e]], x[to_i[e]], y[to_i[e]]))
    }

    has_tooltip <- "tooltip" %in% names(nodes)
    for (i in seq_len(n)) {
        label <- xml_escape(as.character(nodes$label[i]))
        tip_raw <- if (has_tooltip && !is.na(nodes$tooltip[i]) &&
            nzchar(nodes$tooltip[i])) {
            nodes$tooltip[i]
        } else {
            nodes$label[i]
        }
        tip <- xml_escape(as.character(tip_raw))
        # Preserve newlines in tooltip via XML numeric entity so the
        # SVG stays on one line on disk; browsers render &#10; as a
        # line break in native tooltip rendering.
        tip <- gsub("\n", "&#10;", tip, fixed = TRUE)
        node <- sprintf(paste0(
                               '<circle class="node" cx="%.1f" cy="%.1f" r="6"><title>%s</title></circle>',
                               '<text class="label" x="%.1f" y="%.1f">%s</text>'),
                        x[i], y[i], tip,
                        x[i] + 8, y[i] + 4, label)
        href <- nodes$href[i]
        if (!is.na(href) && nzchar(href)) {
            node <- sprintf('<a xlink:href="%s">%s</a>', xml_escape(href), node)
        }
        out <- c(out, node)
    }

    c(out, "</svg>")
}

#' Escape text for safe inclusion in SVG/XML
#' @noRd
xml_escape <- function(s) {
    s <- gsub("&", "&amp;", s, fixed = TRUE)
    s <- gsub("<", "&lt;", s, fixed = TRUE)
    s <- gsub(">", "&gt;", s, fixed = TRUE)
    s <- gsub('"', "&quot;", s, fixed = TRUE)
    s
}

# Tests for graph_svg.R

library(saber)

edges <- data.frame(from = c("a", "a"), to = c("b", "c"),
                    stringsAsFactors = FALSE)
svg <- graph_svg(edges)

expect_true(is.character(svg))
expect_true(any(grepl("^<svg", svg)))
expect_true(any(grepl("</svg>$", svg)))
expect_equal(sum(grepl("<circle ", svg)), 3L)
expect_equal(sum(grepl("<line ", svg)), 2L)
expect_true(any(grepl("<title>a</title>", svg)))

# Deterministic layout
svg1 <- graph_svg(edges, seed = 42L)
svg2 <- graph_svg(edges, seed = 42L)
expect_equal(svg1, svg2)

svg3 <- graph_svg(edges, seed = 99L)
expect_false(identical(svg1, svg3))

# Hrefs wrap in anchors
nodes <- data.frame(id = c("a", "b", "c"),
                    label = c("Alpha", "Beta", "Gamma"),
                    href = c("a.html", "b.html", NA_character_),
                    stringsAsFactors = FALSE)
svg <- graph_svg(edges, nodes)
expect_true(any(grepl('xlink:href="a.html"', svg, fixed = TRUE)))
expect_equal(sum(grepl("<a ", svg)), 2L)

# XML escaping
edges2 <- data.frame(from = "x<y", to = "z&w", stringsAsFactors = FALSE)
svg <- graph_svg(edges2)
expect_true(any(grepl("<title>x&lt;y</title>", svg)))
expect_true(any(grepl("<title>z&amp;w</title>", svg)))

# Custom tooltip column overrides label-based tooltip
nodes3 <- data.frame(id = c("a", "b", "c"),
                     label = c("a", "b", "c"),
                     href = NA_character_,
                     tooltip = c("Alpha node\nversion 1.0",
                                 "Beta node\nversion 2.0",
                                 NA_character_),
                     stringsAsFactors = FALSE)
svg <- graph_svg(edges, nodes3)
expect_true(any(grepl("Alpha node", svg, fixed = TRUE)))
expect_true(any(grepl("version 1.0", svg, fixed = TRUE)))
# NA tooltip falls back to label
expect_true(any(grepl("<title>c</title>", svg)))

# Unknown node id -> clear error
expect_error(graph_svg(edges,
                       nodes = data.frame(id = c("a", "b"), label = c("a", "b"),
                                          href = NA_character_,
                                          stringsAsFactors = FALSE)),
             "unknown node id")

# Empty edges
svg <- graph_svg(edges = data.frame(from = character(), to = character(),
                                    stringsAsFactors = FALSE),
                 nodes = data.frame(id = c("solo1", "solo2"),
                                    label = c("solo1", "solo2"),
                                    href = NA_character_,
                                    stringsAsFactors = FALSE))
expect_equal(sum(grepl("<line ", svg)), 0L)
expect_equal(sum(grepl("<circle ", svg)), 2L)

# Tests for session-start hook script

scan_dir <- file.path(tempdir(), "test_session_start")
pkg_dir <- file.path(scan_dir, "hookpkg")
sub_dir <- file.path(pkg_dir, "R")
dir.create(sub_dir, recursive = TRUE, showWarnings = FALSE)

writeLines(c(
    "Package: hookpkg",
    "Title: Hook Package",
    "Version: 0.1.0"
), file.path(pkg_dir, "DESCRIPTION"))
writeLines("hook_fn <- function() NULL", file.path(sub_dir, "hook.R"))

system2("git", c("-C", pkg_dir, "init", "-q"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.email", "test@test.com"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.name", "Test"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "add", "-A"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "commit", "-q", "-m", "init"),
        stdout = FALSE, stderr = FALSE)

script <- system.file("scripts", "session-start.R", package = "saber")
old_wd <- getwd()
on.exit(setwd(old_wd), add = TRUE)
setwd(sub_dir)

output <- system2(file.path(R.home("bin"), "Rscript"), c(script, "claude"),
                  stdout = TRUE, stderr = TRUE)

expect_true(length(output) > 0L)
expect_true(identical(trimws(output[[1L]]), "{"))
expect_true(any(grepl('"hookEventName": "SessionStart"', output, fixed = TRUE)))
expect_true(any(grepl('"additionalContext": "# Briefing: hookpkg\\\\n',
                      output)))
expect_false(any(grepl("^# Briefing: hookpkg$", output)))

unlink(scan_dir, recursive = TRUE)

# Tests for session-start hook script.
# Exercises the hook end-to-end via a child Rscript with HOME / CODEX_HOME
# redirected. The system2(env =) plumbing is not honored consistently on
# Windows (env vars get passed as positional args to Rscript), so this is
# gated to dev runs only.
if (!at_home()) exit_file("Session-start hook is *nix dev-only.")

scan_dir <- file.path(tempdir(), "test_session_start")
pkg_dir <- file.path(scan_dir, "hookpkg")
sub_dir <- file.path(pkg_dir, "R")
dir.create(sub_dir, recursive = TRUE, showWarnings = FALSE)

writeLines(c(
    "Package: hookpkg",
    "Title: Hook Package",
    "Version: 0.1.0"
), file.path(pkg_dir, "DESCRIPTION"))
writeLines("hook_fn <- function() NULL", file.path(sub_dir, "hook.R"))

system2("git", c("-C", pkg_dir, "init", "-q"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.email", "test@test.com"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "config", "user.name", "Test"),
        stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "add", "-A"), stdout = FALSE, stderr = FALSE)
system2("git", c("-C", pkg_dir, "commit", "-q", "-m", "init"),
        stdout = FALSE, stderr = FALSE)

script <- system.file("scripts", "session-start.R", package = "saber")
old_wd <- getwd()
on.exit(setwd(old_wd), add = TRUE)
setwd(sub_dir)

home_dir <- file.path(scan_dir, "home")
codex_home <- file.path(scan_dir, "codex_home")
dir.create(file.path(home_dir, ".config", "agents"), recursive = TRUE,
           showWarnings = FALSE)
dir.create(file.path(codex_home, "memories"), recursive = TRUE,
           showWarnings = FALSE)
writeLines(c(
    "# Global Development Preferences",
    "",
    "- Use saber before guessing."
), file.path(home_dir, ".config", "agents", "GLOBAL.md"))
writeLines("saber is meant to be reciprocal",
           file.path(codex_home, "memories", "reciprocal.md"))
memory_dir <- file.path(home_dir, ".claude", "projects",
                        "-home-test-hookpkg", "memory")
dir.create(memory_dir, recursive = TRUE, showWarnings = FALSE)
writeLines(c(
    "- [Memory body](memory-body.md) - hookpkg memory index entry"
), file.path(memory_dir, "MEMORY.md"))
writeLines("This body file should not be preloaded.",
           file.path(memory_dir, "memory-body.md"))

output <- system2(file.path(R.home("bin"), "Rscript"), c(script, "claude"),
                  stdout = TRUE, stderr = TRUE,
                  env = c(sprintf("HOME=%s", home_dir),
                          sprintf("CODEX_HOME=%s", codex_home)))

expect_true(length(output) > 0L)
expect_true(identical(trimws(output[[1L]]), "{"))
expect_true(any(grepl('"hookEventName": "SessionStart"', output, fixed = TRUE)))
expect_true(any(grepl('"additionalContext": "# Briefing: hookpkg\\\\n',
                      output)))
expect_true(any(grepl("## Global Preferences\\n\\n# Global Development Preferences",
                      output, fixed = TRUE)))
expect_true(any(grepl("Use saber before guessing.", output, fixed = TRUE)))
expect_false(any(grepl("hookpkg memory index entry", output, fixed = TRUE)))
expect_true(any(grepl("saber is meant to be reciprocal", output,
                      fixed = TRUE)))
expect_false(any(grepl("^# Briefing: hookpkg$", output)))

codex_output <- system2(file.path(R.home("bin"), "Rscript"), c(script, "codex"),
                        stdout = TRUE, stderr = TRUE,
                        env = c(sprintf("HOME=%s", home_dir),
                                sprintf("CODEX_HOME=%s", codex_home)))

expect_true(any(grepl("## Memory", codex_output, fixed = TRUE)))
expect_true(any(grepl("hookpkg memory index entry", codex_output, fixed = TRUE)))
expect_false(any(grepl("saber is meant to be reciprocal", codex_output,
                       fixed = TRUE)))
expect_false(any(grepl("This body file should not be preloaded.",
                       codex_output, fixed = TRUE)))

corteza_output <- system2(file.path(R.home("bin"), "Rscript"),
                          c(script, "corteza"),
                          stdout = TRUE, stderr = TRUE,
                          env = c(sprintf("HOME=%s", home_dir),
                                  sprintf("CODEX_HOME=%s", codex_home)))

expect_true(any(grepl("hookpkg memory index entry", corteza_output,
                      fixed = TRUE)))
expect_true(any(grepl("saber is meant to be reciprocal", corteza_output,
                      fixed = TRUE)))

unlink(scan_dir, recursive = TRUE)

% tinyrox says don't edit this manually, but it can't stop you!
\name{symbols}
\alias{symbols}
\title{Code intelligence: AST symbol index}
\usage{
symbols(project_dir,
        cache_dir = file.path(tools::R_user_dir("saber", "cache"), "symbols"))
}
\arguments{
\item{project_dir}{Path to the project directory.}

\item{cache_dir}{Directory for symbol cache files.}
}
\value{
A list with components:
  \describe{
    \item{defs}{data.frame(name, file, line, exported)}
    \item{calls}{data.frame(caller, callee, file, line)}
  }
}
\description{
Parse R source files into structured function definitions and
  call relationships.
}
\examples{
# Create a minimal project with R source files
d <- file.path(tempdir(), "demopkg")
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
writeLines("double <- function(x) add(x, x)", file.path(d, "R", "double.R"))

idx <- symbols(d, cache_dir = tempdir())
idx$defs   # function definitions
idx$calls  # call relationships (double calls add)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{graph_svg}
\alias{graph_svg}
\title{Force-directed graph rendering}
\usage{
graph_svg(edges, nodes = NULL, width = 1200L, height = 900L, iterations = 50L,
          seed = 1L)
}
\arguments{
\item{edges}{Data frame with \code{from} and \code{to} columns
holding node ids.}

\item{nodes}{Optional data frame with \code{id}, \code{label},
\code{href} columns. Optionally a \code{tooltip} column (plain
text, may contain newlines) that overrides the default hover
text (which is the label). \code{id} must cover every node
mentioned in \code{edges}. If \code{NULL}, ids from \code{edges}
are used as labels with no hrefs or tooltips.}

\item{width}{SVG viewport width in pixels.}

\item{height}{SVG viewport height in pixels.}

\item{iterations}{Force-simulation steps. 50 is usually enough.}

\item{seed}{Integer seed for the random initial layout (output is
deterministic given the same seed).}
}
\value{
Character vector, one SVG element per line. Write with
  \code{writeLines()}.
}
\description{
Render a graph as static, interactive SVG using a base R
  Fruchterman-Reingold force simulation.
Render a graph as static SVG

Runs a Fruchterman-Reingold force simulation to lay out nodes, then
emits SVG with baked coordinates. Output is interactive via native
browser features: hover tooltips from \code{<title>} elements, click
navigation from \code{<a xlink:href>} wrappers, CSS \code{:hover}
highlighting. No JavaScript.

Suitable for graphs up to roughly a few hundred nodes. The repulsion
step is vectorized via \code{outer()} but allocates an
\code{n x n} matrix per iteration; larger graphs should pre-filter.
}
\examples{
edges <- data.frame(from = c("a", "a", "b"),
                    to = c("b", "c", "c"))
svg <- graph_svg(edges)
writeLines(svg, tempfile(fileext = ".svg"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{blast_radius}
\alias{blast_radius}
\title{Blast radius analysis}
\usage{
blast_radius(fn, project = NULL, include = "r", scan_dir = path.expand("~"),
             cache_dir = file.path(tools::R_user_dir("saber", "cache"), "symbols"),
             exclude = default_exclude())
}
\arguments{
\item{fn}{Character. Function name to search for.}

\item{project}{Character. Project name (or path to project directory).}

\item{include}{Character vector. Any of \code{"r"} (R source, default),
\code{"examples"} (roxygen \verb{@examples} blocks in the target
project), and \code{"vignettes"} (code chunks in the target project's
vignettes).}

\item{scan_dir}{Directory to scan for downstream projects.}

\item{cache_dir}{Directory for symbol cache files.}

\item{exclude}{Character vector of directory basenames to skip when
scanning for downstream projects.}
}
\value{
A data.frame with columns: caller, project, file, line, source.
  \code{source} is one of \code{"r"}, \code{"example"}, \code{"vignette"}.
}
\description{
Find all callers of a function across projects.
Find callers of a function across projects

Given a function name and project, finds all internal callers within that
project and all callers in downstream projects (projects whose DESCRIPTION
lists this one in Depends, Imports, or LinkingTo).

With \code{include = c("r", "examples", "vignettes")} the search can be
extended to references in the target project's roxygen \verb{@examples}
blocks and vignette code chunks (Rmd, qmd, Rnw). Documentation scanning is
target-project only; it does not walk downstream projects' docs.
}
\examples{
# Create a minimal project
d <- file.path(tempdir(), "blastpkg")
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines("helper <- function(x) x + 1", file.path(d, "R", "helper.R"))
writeLines("main <- function(x) helper(x * 2)", file.path(d, "R", "main.R"))

# Find all callers of helper()
blast_radius("helper", project = d, scan_dir = tempdir(),
             cache_dir = tempdir())

# Include roxygen @examples and vignettes from the target project
blast_radius("helper", project = d, include = c("r", "examples", "vignettes"),
             scan_dir = tempdir(), cache_dir = tempdir())
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{agent_context}
\alias{agent_context}
\title{Agent context assembly}
\usage{
agent_context(agent = NULL, project_dir = getwd(), workspace_dir = NULL,
              memory_base = file.path(path.expand("~"), ".claude", "projects"),
              claude_global_path = file.path(path.expand("~"), ".claude", "CLAUDE.md"),
              include_memory = NULL, include_project = NULL,
              include_global = NULL, include_soul = NULL,
              max_memory_lines = 100L)
}
\arguments{
\item{agent}{Consumer identifier: \code{"claude"}, \code{"codex"},
\code{"corteza"}, or \code{NULL} (interactive / unknown). The legacy
\code{"llamar"} identifier is accepted as an alias for \code{"corteza"}.}

\item{project_dir}{Project directory to scan for CLAUDE.md / AGENTS.md.}

\item{workspace_dir}{Optional directory containing SOUL.md and USER.md
(e.g. \code{~/.corteza/workspace}). If \code{NULL}, those files are
skipped.}

\item{memory_base}{Base directory for 'Claude Code' project memory files.}

\item{claude_global_path}{Path to the global 'Claude Code' instructions
file. Defaults to \code{~/.claude/CLAUDE.md}.}

\item{include_memory}{Override default for memory inclusion. Use
\code{TRUE}/\code{FALSE} to force, or \code{NULL} for the agent default.}

\item{include_project}{Override default for project instructions
(CLAUDE.md / AGENTS.md).}

\item{include_global}{Override default for global instructions
(~/.claude/CLAUDE.md / USER.md).}

\item{include_soul}{Override default for SOUL.md inclusion.}

\item{max_memory_lines}{Maximum lines to include from each memory source.}
}
\value{
Character string of assembled context, or empty string if no
  context applies.
}
\description{
Assemble context from memory, instructions, and identity files
  for AI coding agents.
Load context files for an AI coding agent

Returns assembled context (memory, project/global instructions, agent
identity files) tailored to a specific consumer. Files autoloaded by the
agent natively are skipped to avoid duplication.

Defaults per agent:
\itemize{
  \item \code{"claude"} - skips Claude Code project memory and CLAUDE.md
    files (autoloaded by 'Claude Code'). Loads Codex memories, AGENTS.md,
    USER.md, and SOUL.md when present.
  \item \code{"codex"} - skips AGENTS.md and Codex memories (autoloaded by
    Codex). Loads Claude Code project memory, CLAUDE.md, USER.md, and
    SOUL.md when present.
  \item \code{"corteza"} or \code{NULL} - loads everything available.
}

Project and global instructions are resolved by trying both naming
conventions and picking the file relevant to the consumer:
\itemize{
  \item Project: \code{CLAUDE.md} or \code{AGENTS.md}
  \item Global: \code{~/.claude/CLAUDE.md} or
    \code{<workspace_dir>/USER.md}
}

Override the defaults with the \code{include_*} parameters.
}
\examples{
\donttest{
# Codex agent in current project
saber::agent_context(agent = "codex")

# Corteza with workspace files
saber::agent_context(agent = "corteza",
                     workspace_dir = "~/.corteza/workspace")

# Force-include memory regardless of agent default
saber::agent_context(agent = "claude", include_memory = TRUE)
}
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{default_exclude}
\alias{default_exclude}
\title{Default directories to exclude when scanning for projects}
\usage{
default_exclude()
}
\value{
Character vector of directory basenames.
}
\description{
Returns a character vector of directory basenames that are skipped
when scanning for downstream projects. Override by passing a custom
\code{exclude} vector to \code{\link{blast_radius}}.
}
\examples{
default_exclude()
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{projects}
\alias{projects}
\title{Project discovery}
\usage{
projects(scan_dir = path.expand("~"), exclude = default_exclude())
}
\arguments{
\item{scan_dir}{Directory to scan for project directories.}

\item{exclude}{Character vector of directory basenames to skip.}
}
\value{
A data.frame with columns: package, title, version, path, depends,
  imports.
}
\description{
Discover R package projects and map their dependency
  relationships.
Discover R package projects

Scans a directory for subdirectories containing a DESCRIPTION file and
returns their metadata.
}
\examples{
d <- file.path(tempdir(), "scandir")
dir.create(d, showWarnings = FALSE)
pkg <- file.path(d, "mypkg")
dir.create(pkg, showWarnings = FALSE)
writeLines(c("Package: mypkg", "Title: Demo", "Version: 0.1.0"),
           file.path(pkg, "DESCRIPTION"))
projects(scan_dir = d)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{pkg_help}
\alias{pkg_help}
\title{Get help for a package topic as markdown}
\usage{
pkg_help(topic, package, format = c("md", "hugo"))
}
\arguments{
\item{topic}{Character. The help topic name.}

\item{package}{Character. Package name.}

\item{format}{Character. Output format: \code{"md"} (default) for plain
markdown, or \code{"hugo"} for Hugo-compatible markdown with YAML front
matter.}
}
\value{
Character string of markdown help text.
}
\description{
Extracts help documentation and converts it to clean markdown.
}
\examples{
cat(pkg_help("md5sum", "tools"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{pkg_internals}
\alias{pkg_internals}
\title{List internal (non-exported) functions of a package}
\usage{
pkg_internals(package, pattern = NULL)
}
\arguments{
\item{package}{Character. Package name.}

\item{pattern}{Optional regex to filter function names.}
}
\value{
A data.frame with columns: name, args.
}
\description{
Returns functions defined in a package namespace but not exported.
}
\examples{
pkg_internals("tools", pattern = "^check")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{fn_graph}
\alias{fn_graph}
\title{Code intelligence: function call graph}
\usage{
fn_graph(project_dir, include_external = FALSE, ...)
}
\arguments{
\item{project_dir}{Path to the project directory (an R package root).}

\item{include_external}{If \code{TRUE}, also include nodes for
functions called from other packages. Default \code{FALSE}.}

\item{...}{Passed through to \code{\link{graph_svg}} (e.g.,
\code{width}, \code{height}, \code{iterations}, \code{seed}).}
}
\value{
Character vector of SVG lines. Write with \code{writeLines()}.
}
\description{
Render a project's internal function call graph as interactive
  SVG.
Render a function call graph for an R project

Pulls the AST symbol index via \code{\link{symbols}} and renders
internal function-to-function call edges as SVG. External calls
(into other packages) are dropped by default.
}
\examples{
d <- file.path(tempdir(), "fngdemo")
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines(c("Package: demo", "Version: 0.1.0"),
           file.path(d, "DESCRIPTION"))
writeLines("add <- function(x, y) x + y", file.path(d, "R", "add.R"))
writeLines("double <- function(x) add(x, x)",
           file.path(d, "R", "double.R"))
svg <- fn_graph(d)
writeLines(svg, tempfile(fileext = ".svg"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{briefing}
\alias{briefing}
\title{Project briefings}
\usage{
briefing(project = NULL, scan_dir = path.expand("~"),
         briefs_dir = file.path(tools::R_user_dir("saber", "cache"), "briefs"))
}
\arguments{
\item{project}{Project name. If NULL, inferred from the current working
directory basename.}

\item{scan_dir}{Directory to scan for project directories.}

\item{briefs_dir}{Directory to write briefing markdown files.}
}
\value{
The briefing text (character string), returned invisibly. Emitted
  via \code{message()} and written to \code{briefs_dir/{project}.md}.
}
\description{
Generate project context for AI coding agents.
Generate a project briefing

Produces a concise markdown briefing combining DESCRIPTION metadata,
downstream dependents, and recent git commits. Written to the user
cache directory so both the agent and user see the same context.

For runtime context (memory, identity files, project instructions),
see \code{\link{agent_context}}.
}
\examples{
d <- file.path(tempdir(), "briefpkg")
dir.create(file.path(d, "R"), recursive = TRUE, showWarnings = FALSE)
writeLines(c("Package: briefpkg", "Title: Demo", "Version: 0.1.0"),
           file.path(d, "DESCRIPTION"))
briefing("briefpkg", scan_dir = tempdir(),
         briefs_dir = file.path(tempdir(), "briefs"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{find_downstream}
\alias{find_downstream}
\title{Find projects that depend on a given package}
\usage{
find_downstream(package, scan_dir = path.expand("~"),
                exclude = default_exclude())
}
\arguments{
\item{package}{Character. Package name to search for.}

\item{scan_dir}{Directory to scan for project directories.}

\item{exclude}{Character vector of directory basenames to skip.}
}
\value{
Character vector of project names that depend on \code{package}.
}
\description{
Scans DESCRIPTION files in project directories under \code{scan_dir}
for Depends, Imports, or LinkingTo fields that reference \code{package}.
}
\examples{
d <- file.path(tempdir(), "dsdir")
dir.create(d, showWarnings = FALSE)
pkg <- file.path(d, "child")
dir.create(pkg, showWarnings = FALSE)
writeLines(c("Package: child", "Version: 0.1.0", "Imports: parent"),
           file.path(pkg, "DESCRIPTION"))
find_downstream("parent", scan_dir = d)
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{pkg_exports}
\alias{pkg_exports}
\title{Package introspection}
\usage{
pkg_exports(package, pattern = NULL)
}
\arguments{
\item{package}{Character. Package name.}

\item{pattern}{Optional regex to filter function names.}
}
\value{
A data.frame with columns: name, args.
}
\description{
Query installed R packages for exported functions, internal
  functions, and help documentation.
List exported functions of a package

Returns a data.frame of exported functions with their argument signatures.
}
\examples{
pkg_exports("tools")
pkg_exports("tools", pattern = "^Rd")
}

% tinyrox says don't edit this manually, but it can't stop you!
\name{pkg_graph}
\alias{pkg_graph}
\title{Project discovery: package dependency graph}
\usage{
pkg_graph(scan_dir = path.expand("~"), packages = NULL,
          include_suggests = FALSE, ...)
}
\arguments{
\item{scan_dir}{Directory to scan for project directories.}

\item{packages}{Optional character vector limiting the graph to
these packages.}

\item{include_suggests}{If \code{TRUE}, also include edges for
packages in each project's \code{Suggests} field. Default
\code{FALSE} (only \code{Depends} and \code{Imports}).}

\item{...}{Passed through to \code{\link{graph_svg}}.}
}
\value{
Character vector of SVG lines. Write with \code{writeLines()}.
}
\description{
Render the dependency graph across a set of R packages as
  interactive SVG.
Render a package-level dependency graph

Discovers R packages under \code{scan_dir} via \code{\link{projects}},
parses each one's \code{Depends} and \code{Imports} fields, and
renders edges between packages that both live in \code{scan_dir}.
External CRAN dependencies are dropped.
}
\examples{
d <- file.path(tempdir(), "pkgdemo")
dir.create(file.path(d, "parent"), recursive = TRUE, showWarnings = FALSE)
dir.create(file.path(d, "child"), showWarnings = FALSE)
writeLines(c("Package: parent", "Title: P", "Version: 0.1.0"),
           file.path(d, "parent", "DESCRIPTION"))
writeLines(c("Package: child", "Title: C", "Version: 0.1.0",
             "Imports: parent"),
           file.path(d, "child", "DESCRIPTION"))
svg <- pkg_graph(scan_dir = d)
writeLines(svg, tempfile(fileext = ".svg"))
}

% tinyrox says don't edit this manually, but it can't stop you!
\docType{package}
\name{saber-package}
\alias{saber}
\alias{saber-package}
\title{saber: Context Engineering for Large Language Model Agents}
\description{
Context-engineering primitives for Artificial Intelligence (AI) coding
agents working in R.
Assembles agent context from memory and instruction files, traces
function call blast radius across projects, generates project
briefings, parses R source into structured Abstract Syntax Tree
symbol indices, discovers dependency graphs, and introspects
installed packages.
}
\author{
\strong{Maintainer}: Troy Hernandez \email{troy@cornball.ai}
}
\keyword{internal}
